MoE Kernel (#2465)

* add moe grouped gemm kernel

* add benchmark, README

* remove formatting from __init__.py
This commit is contained in:
jeromeku 2025-05-02 20:59:23 -07:00 committed by GitHub
commit b7fc12c8be
24 changed files with 6584 additions and 1 deletions

175
.gitignore vendored Normal file
View file

@ -0,0 +1,175 @@
# Byte-compiled / optimized / DLL files
__pycache__/
*.py[cod]
*.class
# C extensions
*.so
# Distribution / packaging
.Python
build/
develop-eggs/
dist/
downloads/
eggs/
.eggs/
lib/
lib64/
parts/
sdist/
var/
wheels/
share/python-wheels/
*.egg-info/
.installed.cfg
*.egg
MANIFEST
# PyInstaller
# Usually these files are written by a python script from a template
# before PyInstaller builds the exe, so as to inject date/other infos into it.
*.manifest
*.spec
# Installer logs
pip-log.txt
pip-delete-this-directory.txt
# Unit test / coverage reports
htmlcov/
.tox/
.nox/
.coverage
.coverage.*
.cache
nosetests.xml
coverage.xml
*.cover
*.py,cover
.hypothesis/
.pytest_cache/
cover/
# Translations
*.mo
*.pot
# Django stuff:
*.log
local_settings.py
db.sqlite3
db.sqlite3-journal
# Flask stuff:
instance/
.webassets-cache
# Scrapy stuff:
.scrapy
# Sphinx documentation
docs/_build/
# PyBuilder
.pybuilder/
target/
# Jupyter Notebook
.ipynb_checkpoints
# IPython
profile_default/
ipython_config.py
# pyenv
# For a library or package, you might want to ignore these files since the code is
# intended to run in multiple environments; otherwise, check them in:
# .python-version
# pipenv
# According to pypa/pipenv#598, it is recommended to include Pipfile.lock in version control.
# However, in case of collaboration, if having platform-specific dependencies or dependencies
# having no cross-platform support, pipenv may install dependencies that don't work, or not
# install all needed dependencies.
#Pipfile.lock
# UV
# Similar to Pipfile.lock, it is generally recommended to include uv.lock in version control.
# This is especially recommended for binary packages to ensure reproducibility, and is more
# commonly ignored for libraries.
#uv.lock
# poetry
# Similar to Pipfile.lock, it is generally recommended to include poetry.lock in version control.
# This is especially recommended for binary packages to ensure reproducibility, and is more
# commonly ignored for libraries.
# https://python-poetry.org/docs/basic-usage/#commit-your-poetrylock-file-to-version-control
#poetry.lock
# pdm
# Similar to Pipfile.lock, it is generally recommended to include pdm.lock in version control.
#pdm.lock
# pdm stores project-wide configurations in .pdm.toml, but it is recommended to not include it
# in version control.
# https://pdm.fming.dev/latest/usage/project/#working-with-version-control
.pdm.toml
.pdm-python
.pdm-build/
# PEP 582; used by e.g. github.com/David-OConnor/pyflow and github.com/pdm-project/pdm
__pypackages__/
# Celery stuff
celerybeat-schedule
celerybeat.pid
# SageMath parsed files
*.sage.py
# Environments
.env
.venv
env/
venv/
ENV/
env.bak/
venv.bak/
# Spyder project settings
.spyderproject
.spyproject
# Rope project settings
.ropeproject
# mkdocs documentation
/site
# mypy
.mypy_cache/
.dmypy.json
dmypy.json
# Pyre type checker
.pyre/
# pytype static type analyzer
.pytype/
# Cython debug symbols
cython_debug/
# PyCharm
# JetBrains specific template is maintained in a separate JetBrains.gitignore that can
# be found at https://github.com/github/gitignore/blob/main/Global/JetBrains.gitignore
# and can be added to the global gitignore or merged into this file. For a more nuclear
# option (not recommended) you can uncomment the following to ignore the entire idea folder.
#.idea/
# Ruff stuff:
.ruff_cache/
# PyPI configuration file
.pypirc
.vscode

View file

@ -225,4 +225,4 @@ from .tokenizer_utils import *
from .trainer import *
# Patch TRL trainers for backwards compatibility
_patch_trl_trainer()
_patch_trl_trainer()

661
unsloth/kernels/moe/LICENSE Normal file
View file

@ -0,0 +1,661 @@
GNU AFFERO GENERAL PUBLIC LICENSE
Version 3, 19 November 2007
Copyright (C) 2007 Free Software Foundation, Inc. <https://fsf.org/>
Everyone is permitted to copy and distribute verbatim copies
of this license document, but changing it is not allowed.
Preamble
The GNU Affero General Public License is a free, copyleft license for
software and other kinds of works, specifically designed to ensure
cooperation with the community in the case of network server software.
The licenses for most software and other practical works are designed
to take away your freedom to share and change the works. By contrast,
our General Public Licenses are intended to guarantee your freedom to
share and change all versions of a program--to make sure it remains free
software for all its users.
When we speak of free software, we are referring to freedom, not
price. Our General Public Licenses are designed to make sure that you
have the freedom to distribute copies of free software (and charge for
them if you wish), that you receive source code or can get it if you
want it, that you can change the software or use pieces of it in new
free programs, and that you know you can do these things.
Developers that use our General Public Licenses protect your rights
with two steps: (1) assert copyright on the software, and (2) offer
you this License which gives you legal permission to copy, distribute
and/or modify the software.
A secondary benefit of defending all users' freedom is that
improvements made in alternate versions of the program, if they
receive widespread use, become available for other developers to
incorporate. Many developers of free software are heartened and
encouraged by the resulting cooperation. However, in the case of
software used on network servers, this result may fail to come about.
The GNU General Public License permits making a modified version and
letting the public access it on a server without ever releasing its
source code to the public.
The GNU Affero General Public License is designed specifically to
ensure that, in such cases, the modified source code becomes available
to the community. It requires the operator of a network server to
provide the source code of the modified version running there to the
users of that server. Therefore, public use of a modified version, on
a publicly accessible server, gives the public access to the source
code of the modified version.
An older license, called the Affero General Public License and
published by Affero, was designed to accomplish similar goals. This is
a different license, not a version of the Affero GPL, but Affero has
released a new version of the Affero GPL which permits relicensing under
this license.
The precise terms and conditions for copying, distribution and
modification follow.
TERMS AND CONDITIONS
0. Definitions.
"This License" refers to version 3 of the GNU Affero General Public License.
"Copyright" also means copyright-like laws that apply to other kinds of
works, such as semiconductor masks.
"The Program" refers to any copyrightable work licensed under this
License. Each licensee is addressed as "you". "Licensees" and
"recipients" may be individuals or organizations.
To "modify" a work means to copy from or adapt all or part of the work
in a fashion requiring copyright permission, other than the making of an
exact copy. The resulting work is called a "modified version" of the
earlier work or a work "based on" the earlier work.
A "covered work" means either the unmodified Program or a work based
on the Program.
To "propagate" a work means to do anything with it that, without
permission, would make you directly or secondarily liable for
infringement under applicable copyright law, except executing it on a
computer or modifying a private copy. Propagation includes copying,
distribution (with or without modification), making available to the
public, and in some countries other activities as well.
To "convey" a work means any kind of propagation that enables other
parties to make or receive copies. Mere interaction with a user through
a computer network, with no transfer of a copy, is not conveying.
An interactive user interface displays "Appropriate Legal Notices"
to the extent that it includes a convenient and prominently visible
feature that (1) displays an appropriate copyright notice, and (2)
tells the user that there is no warranty for the work (except to the
extent that warranties are provided), that licensees may convey the
work under this License, and how to view a copy of this License. If
the interface presents a list of user commands or options, such as a
menu, a prominent item in the list meets this criterion.
1. Source Code.
The "source code" for a work means the preferred form of the work
for making modifications to it. "Object code" means any non-source
form of a work.
A "Standard Interface" means an interface that either is an official
standard defined by a recognized standards body, or, in the case of
interfaces specified for a particular programming language, one that
is widely used among developers working in that language.
The "System Libraries" of an executable work include anything, other
than the work as a whole, that (a) is included in the normal form of
packaging a Major Component, but which is not part of that Major
Component, and (b) serves only to enable use of the work with that
Major Component, or to implement a Standard Interface for which an
implementation is available to the public in source code form. A
"Major Component", in this context, means a major essential component
(kernel, window system, and so on) of the specific operating system
(if any) on which the executable work runs, or a compiler used to
produce the work, or an object code interpreter used to run it.
The "Corresponding Source" for a work in object code form means all
the source code needed to generate, install, and (for an executable
work) run the object code and to modify the work, including scripts to
control those activities. However, it does not include the work's
System Libraries, or general-purpose tools or generally available free
programs which are used unmodified in performing those activities but
which are not part of the work. For example, Corresponding Source
includes interface definition files associated with source files for
the work, and the source code for shared libraries and dynamically
linked subprograms that the work is specifically designed to require,
such as by intimate data communication or control flow between those
subprograms and other parts of the work.
The Corresponding Source need not include anything that users
can regenerate automatically from other parts of the Corresponding
Source.
The Corresponding Source for a work in source code form is that
same work.
2. Basic Permissions.
All rights granted under this License are granted for the term of
copyright on the Program, and are irrevocable provided the stated
conditions are met. This License explicitly affirms your unlimited
permission to run the unmodified Program. The output from running a
covered work is covered by this License only if the output, given its
content, constitutes a covered work. This License acknowledges your
rights of fair use or other equivalent, as provided by copyright law.
You may make, run and propagate covered works that you do not
convey, without conditions so long as your license otherwise remains
in force. You may convey covered works to others for the sole purpose
of having them make modifications exclusively for you, or provide you
with facilities for running those works, provided that you comply with
the terms of this License in conveying all material for which you do
not control copyright. Those thus making or running the covered works
for you must do so exclusively on your behalf, under your direction
and control, on terms that prohibit them from making any copies of
your copyrighted material outside their relationship with you.
Conveying under any other circumstances is permitted solely under
the conditions stated below. Sublicensing is not allowed; section 10
makes it unnecessary.
3. Protecting Users' Legal Rights From Anti-Circumvention Law.
No covered work shall be deemed part of an effective technological
measure under any applicable law fulfilling obligations under article
11 of the WIPO copyright treaty adopted on 20 December 1996, or
similar laws prohibiting or restricting circumvention of such
measures.
When you convey a covered work, you waive any legal power to forbid
circumvention of technological measures to the extent such circumvention
is effected by exercising rights under this License with respect to
the covered work, and you disclaim any intention to limit operation or
modification of the work as a means of enforcing, against the work's
users, your or third parties' legal rights to forbid circumvention of
technological measures.
4. Conveying Verbatim Copies.
You may convey verbatim copies of the Program's source code as you
receive it, in any medium, provided that you conspicuously and
appropriately publish on each copy an appropriate copyright notice;
keep intact all notices stating that this License and any
non-permissive terms added in accord with section 7 apply to the code;
keep intact all notices of the absence of any warranty; and give all
recipients a copy of this License along with the Program.
You may charge any price or no price for each copy that you convey,
and you may offer support or warranty protection for a fee.
5. Conveying Modified Source Versions.
You may convey a work based on the Program, or the modifications to
produce it from the Program, in the form of source code under the
terms of section 4, provided that you also meet all of these conditions:
a) The work must carry prominent notices stating that you modified
it, and giving a relevant date.
b) The work must carry prominent notices stating that it is
released under this License and any conditions added under section
7. This requirement modifies the requirement in section 4 to
"keep intact all notices".
c) You must license the entire work, as a whole, under this
License to anyone who comes into possession of a copy. This
License will therefore apply, along with any applicable section 7
additional terms, to the whole of the work, and all its parts,
regardless of how they are packaged. This License gives no
permission to license the work in any other way, but it does not
invalidate such permission if you have separately received it.
d) If the work has interactive user interfaces, each must display
Appropriate Legal Notices; however, if the Program has interactive
interfaces that do not display Appropriate Legal Notices, your
work need not make them do so.
A compilation of a covered work with other separate and independent
works, which are not by their nature extensions of the covered work,
and which are not combined with it such as to form a larger program,
in or on a volume of a storage or distribution medium, is called an
"aggregate" if the compilation and its resulting copyright are not
used to limit the access or legal rights of the compilation's users
beyond what the individual works permit. Inclusion of a covered work
in an aggregate does not cause this License to apply to the other
parts of the aggregate.
6. Conveying Non-Source Forms.
You may convey a covered work in object code form under the terms
of sections 4 and 5, provided that you also convey the
machine-readable Corresponding Source under the terms of this License,
in one of these ways:
a) Convey the object code in, or embodied in, a physical product
(including a physical distribution medium), accompanied by the
Corresponding Source fixed on a durable physical medium
customarily used for software interchange.
b) Convey the object code in, or embodied in, a physical product
(including a physical distribution medium), accompanied by a
written offer, valid for at least three years and valid for as
long as you offer spare parts or customer support for that product
model, to give anyone who possesses the object code either (1) a
copy of the Corresponding Source for all the software in the
product that is covered by this License, on a durable physical
medium customarily used for software interchange, for a price no
more than your reasonable cost of physically performing this
conveying of source, or (2) access to copy the
Corresponding Source from a network server at no charge.
c) Convey individual copies of the object code with a copy of the
written offer to provide the Corresponding Source. This
alternative is allowed only occasionally and noncommercially, and
only if you received the object code with such an offer, in accord
with subsection 6b.
d) Convey the object code by offering access from a designated
place (gratis or for a charge), and offer equivalent access to the
Corresponding Source in the same way through the same place at no
further charge. You need not require recipients to copy the
Corresponding Source along with the object code. If the place to
copy the object code is a network server, the Corresponding Source
may be on a different server (operated by you or a third party)
that supports equivalent copying facilities, provided you maintain
clear directions next to the object code saying where to find the
Corresponding Source. Regardless of what server hosts the
Corresponding Source, you remain obligated to ensure that it is
available for as long as needed to satisfy these requirements.
e) Convey the object code using peer-to-peer transmission, provided
you inform other peers where the object code and Corresponding
Source of the work are being offered to the general public at no
charge under subsection 6d.
A separable portion of the object code, whose source code is excluded
from the Corresponding Source as a System Library, need not be
included in conveying the object code work.
A "User Product" is either (1) a "consumer product", which means any
tangible personal property which is normally used for personal, family,
or household purposes, or (2) anything designed or sold for incorporation
into a dwelling. In determining whether a product is a consumer product,
doubtful cases shall be resolved in favor of coverage. For a particular
product received by a particular user, "normally used" refers to a
typical or common use of that class of product, regardless of the status
of the particular user or of the way in which the particular user
actually uses, or expects or is expected to use, the product. A product
is a consumer product regardless of whether the product has substantial
commercial, industrial or non-consumer uses, unless such uses represent
the only significant mode of use of the product.
"Installation Information" for a User Product means any methods,
procedures, authorization keys, or other information required to install
and execute modified versions of a covered work in that User Product from
a modified version of its Corresponding Source. The information must
suffice to ensure that the continued functioning of the modified object
code is in no case prevented or interfered with solely because
modification has been made.
If you convey an object code work under this section in, or with, or
specifically for use in, a User Product, and the conveying occurs as
part of a transaction in which the right of possession and use of the
User Product is transferred to the recipient in perpetuity or for a
fixed term (regardless of how the transaction is characterized), the
Corresponding Source conveyed under this section must be accompanied
by the Installation Information. But this requirement does not apply
if neither you nor any third party retains the ability to install
modified object code on the User Product (for example, the work has
been installed in ROM).
The requirement to provide Installation Information does not include a
requirement to continue to provide support service, warranty, or updates
for a work that has been modified or installed by the recipient, or for
the User Product in which it has been modified or installed. Access to a
network may be denied when the modification itself materially and
adversely affects the operation of the network or violates the rules and
protocols for communication across the network.
Corresponding Source conveyed, and Installation Information provided,
in accord with this section must be in a format that is publicly
documented (and with an implementation available to the public in
source code form), and must require no special password or key for
unpacking, reading or copying.
7. Additional Terms.
"Additional permissions" are terms that supplement the terms of this
License by making exceptions from one or more of its conditions.
Additional permissions that are applicable to the entire Program shall
be treated as though they were included in this License, to the extent
that they are valid under applicable law. If additional permissions
apply only to part of the Program, that part may be used separately
under those permissions, but the entire Program remains governed by
this License without regard to the additional permissions.
When you convey a copy of a covered work, you may at your option
remove any additional permissions from that copy, or from any part of
it. (Additional permissions may be written to require their own
removal in certain cases when you modify the work.) You may place
additional permissions on material, added by you to a covered work,
for which you have or can give appropriate copyright permission.
Notwithstanding any other provision of this License, for material you
add to a covered work, you may (if authorized by the copyright holders of
that material) supplement the terms of this License with terms:
a) Disclaiming warranty or limiting liability differently from the
terms of sections 15 and 16 of this License; or
b) Requiring preservation of specified reasonable legal notices or
author attributions in that material or in the Appropriate Legal
Notices displayed by works containing it; or
c) Prohibiting misrepresentation of the origin of that material, or
requiring that modified versions of such material be marked in
reasonable ways as different from the original version; or
d) Limiting the use for publicity purposes of names of licensors or
authors of the material; or
e) Declining to grant rights under trademark law for use of some
trade names, trademarks, or service marks; or
f) Requiring indemnification of licensors and authors of that
material by anyone who conveys the material (or modified versions of
it) with contractual assumptions of liability to the recipient, for
any liability that these contractual assumptions directly impose on
those licensors and authors.
All other non-permissive additional terms are considered "further
restrictions" within the meaning of section 10. If the Program as you
received it, or any part of it, contains a notice stating that it is
governed by this License along with a term that is a further
restriction, you may remove that term. If a license document contains
a further restriction but permits relicensing or conveying under this
License, you may add to a covered work material governed by the terms
of that license document, provided that the further restriction does
not survive such relicensing or conveying.
If you add terms to a covered work in accord with this section, you
must place, in the relevant source files, a statement of the
additional terms that apply to those files, or a notice indicating
where to find the applicable terms.
Additional terms, permissive or non-permissive, may be stated in the
form of a separately written license, or stated as exceptions;
the above requirements apply either way.
8. Termination.
You may not propagate or modify a covered work except as expressly
provided under this License. Any attempt otherwise to propagate or
modify it is void, and will automatically terminate your rights under
this License (including any patent licenses granted under the third
paragraph of section 11).
However, if you cease all violation of this License, then your
license from a particular copyright holder is reinstated (a)
provisionally, unless and until the copyright holder explicitly and
finally terminates your license, and (b) permanently, if the copyright
holder fails to notify you of the violation by some reasonable means
prior to 60 days after the cessation.
Moreover, your license from a particular copyright holder is
reinstated permanently if the copyright holder notifies you of the
violation by some reasonable means, this is the first time you have
received notice of violation of this License (for any work) from that
copyright holder, and you cure the violation prior to 30 days after
your receipt of the notice.
Termination of your rights under this section does not terminate the
licenses of parties who have received copies or rights from you under
this License. If your rights have been terminated and not permanently
reinstated, you do not qualify to receive new licenses for the same
material under section 10.
9. Acceptance Not Required for Having Copies.
You are not required to accept this License in order to receive or
run a copy of the Program. Ancillary propagation of a covered work
occurring solely as a consequence of using peer-to-peer transmission
to receive a copy likewise does not require acceptance. However,
nothing other than this License grants you permission to propagate or
modify any covered work. These actions infringe copyright if you do
not accept this License. Therefore, by modifying or propagating a
covered work, you indicate your acceptance of this License to do so.
10. Automatic Licensing of Downstream Recipients.
Each time you convey a covered work, the recipient automatically
receives a license from the original licensors, to run, modify and
propagate that work, subject to this License. You are not responsible
for enforcing compliance by third parties with this License.
An "entity transaction" is a transaction transferring control of an
organization, or substantially all assets of one, or subdividing an
organization, or merging organizations. If propagation of a covered
work results from an entity transaction, each party to that
transaction who receives a copy of the work also receives whatever
licenses to the work the party's predecessor in interest had or could
give under the previous paragraph, plus a right to possession of the
Corresponding Source of the work from the predecessor in interest, if
the predecessor has it or can get it with reasonable efforts.
You may not impose any further restrictions on the exercise of the
rights granted or affirmed under this License. For example, you may
not impose a license fee, royalty, or other charge for exercise of
rights granted under this License, and you may not initiate litigation
(including a cross-claim or counterclaim in a lawsuit) alleging that
any patent claim is infringed by making, using, selling, offering for
sale, or importing the Program or any portion of it.
11. Patents.
A "contributor" is a copyright holder who authorizes use under this
License of the Program or a work on which the Program is based. The
work thus licensed is called the contributor's "contributor version".
A contributor's "essential patent claims" are all patent claims
owned or controlled by the contributor, whether already acquired or
hereafter acquired, that would be infringed by some manner, permitted
by this License, of making, using, or selling its contributor version,
but do not include claims that would be infringed only as a
consequence of further modification of the contributor version. For
purposes of this definition, "control" includes the right to grant
patent sublicenses in a manner consistent with the requirements of
this License.
Each contributor grants you a non-exclusive, worldwide, royalty-free
patent license under the contributor's essential patent claims, to
make, use, sell, offer for sale, import and otherwise run, modify and
propagate the contents of its contributor version.
In the following three paragraphs, a "patent license" is any express
agreement or commitment, however denominated, not to enforce a patent
(such as an express permission to practice a patent or covenant not to
sue for patent infringement). To "grant" such a patent license to a
party means to make such an agreement or commitment not to enforce a
patent against the party.
If you convey a covered work, knowingly relying on a patent license,
and the Corresponding Source of the work is not available for anyone
to copy, free of charge and under the terms of this License, through a
publicly available network server or other readily accessible means,
then you must either (1) cause the Corresponding Source to be so
available, or (2) arrange to deprive yourself of the benefit of the
patent license for this particular work, or (3) arrange, in a manner
consistent with the requirements of this License, to extend the patent
license to downstream recipients. "Knowingly relying" means you have
actual knowledge that, but for the patent license, your conveying the
covered work in a country, or your recipient's use of the covered work
in a country, would infringe one or more identifiable patents in that
country that you have reason to believe are valid.
If, pursuant to or in connection with a single transaction or
arrangement, you convey, or propagate by procuring conveyance of, a
covered work, and grant a patent license to some of the parties
receiving the covered work authorizing them to use, propagate, modify
or convey a specific copy of the covered work, then the patent license
you grant is automatically extended to all recipients of the covered
work and works based on it.
A patent license is "discriminatory" if it does not include within
the scope of its coverage, prohibits the exercise of, or is
conditioned on the non-exercise of one or more of the rights that are
specifically granted under this License. You may not convey a covered
work if you are a party to an arrangement with a third party that is
in the business of distributing software, under which you make payment
to the third party based on the extent of your activity of conveying
the work, and under which the third party grants, to any of the
parties who would receive the covered work from you, a discriminatory
patent license (a) in connection with copies of the covered work
conveyed by you (or copies made from those copies), or (b) primarily
for and in connection with specific products or compilations that
contain the covered work, unless you entered into that arrangement,
or that patent license was granted, prior to 28 March 2007.
Nothing in this License shall be construed as excluding or limiting
any implied license or other defenses to infringement that may
otherwise be available to you under applicable patent law.
12. No Surrender of Others' Freedom.
If conditions are imposed on you (whether by court order, agreement or
otherwise) that contradict the conditions of this License, they do not
excuse you from the conditions of this License. If you cannot convey a
covered work so as to satisfy simultaneously your obligations under this
License and any other pertinent obligations, then as a consequence you may
not convey it at all. For example, if you agree to terms that obligate you
to collect a royalty for further conveying from those to whom you convey
the Program, the only way you could satisfy both those terms and this
License would be to refrain entirely from conveying the Program.
13. Remote Network Interaction; Use with the GNU General Public License.
Notwithstanding any other provision of this License, if you modify the
Program, your modified version must prominently offer all users
interacting with it remotely through a computer network (if your version
supports such interaction) an opportunity to receive the Corresponding
Source of your version by providing access to the Corresponding Source
from a network server at no charge, through some standard or customary
means of facilitating copying of software. This Corresponding Source
shall include the Corresponding Source for any work covered by version 3
of the GNU General Public License that is incorporated pursuant to the
following paragraph.
Notwithstanding any other provision of this License, you have
permission to link or combine any covered work with a work licensed
under version 3 of the GNU General Public License into a single
combined work, and to convey the resulting work. The terms of this
License will continue to apply to the part which is the covered work,
but the work with which it is combined will remain governed by version
3 of the GNU General Public License.
14. Revised Versions of this License.
The Free Software Foundation may publish revised and/or new versions of
the GNU Affero General Public License from time to time. Such new versions
will be similar in spirit to the present version, but may differ in detail to
address new problems or concerns.
Each version is given a distinguishing version number. If the
Program specifies that a certain numbered version of the GNU Affero General
Public License "or any later version" applies to it, you have the
option of following the terms and conditions either of that numbered
version or of any later version published by the Free Software
Foundation. If the Program does not specify a version number of the
GNU Affero General Public License, you may choose any version ever published
by the Free Software Foundation.
If the Program specifies that a proxy can decide which future
versions of the GNU Affero General Public License can be used, that proxy's
public statement of acceptance of a version permanently authorizes you
to choose that version for the Program.
Later license versions may give you additional or different
permissions. However, no additional obligations are imposed on any
author or copyright holder as a result of your choosing to follow a
later version.
15. Disclaimer of Warranty.
THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
APPLICABLE LAW. EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE. THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
IS WITH YOU. SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
16. Limitation of Liability.
IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
17. Interpretation of Sections 15 and 16.
If the disclaimer of warranty and limitation of liability provided
above cannot be given local legal effect according to their terms,
reviewing courts shall apply local law that most closely approximates
an absolute waiver of all civil liability in connection with the
Program, unless a warranty or assumption of liability accompanies a
copy of the Program in return for a fee.
END OF TERMS AND CONDITIONS
How to Apply These Terms to Your New Programs
If you develop a new program, and you want it to be of the greatest
possible use to the public, the best way to achieve this is to make it
free software which everyone can redistribute and change under these terms.
To do so, attach the following notices to the program. It is safest
to attach them to the start of each source file to most effectively
state the exclusion of warranty; and each file should have at least
the "copyright" line and a pointer to where the full notice is found.
<one line to give the program's name and a brief idea of what it does.>
Copyright (C) <year> <name of author>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.
You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <https://www.gnu.org/licenses/>.
Also add information on how to contact you by electronic and paper mail.
If your software can interact with users remotely through a computer
network, you should also make sure that it provides a way for users to
get its source. For example, if your program is a web application, its
interface could display a "Source" link that leads users to an archive
of the code. There are many ways you could offer source, and different
solutions will be better for different programs; see section 13 for the
specific requirements.
You should also get your employer (if you work as a programmer) or school,
if any, to sign a "copyright disclaimer" for the program, if necessary.
For more information on this, and how to apply and follow the GNU AGPL, see
<https://www.gnu.org/licenses/>.

View file

@ -0,0 +1,72 @@
## MoE Grouped GEMM
Optimized implementation of `MoE MLP Block`.
### Background
`MoE MLP` requires the following steps:
- Calculate `topk_weights` and `topk_indices`
- If using a grouped gemm implementation, calculate permutation indices needed to rearrange tokens grouped by expert
- For each expert:
- `expert_tokens`: gather the tokens assigned to the expert
- `first_gemm`: `gate / up proj` @ `expert_tokens`
- `silu_and_mul`: `silu` and `mul` of `first_gemm`
- `second_gemm`: `silu_and_mul` @ `down proj`
- `scatter_second_gemm`: scatter the `second_gemm` to the original token order
- `topk_weight_mul`: `second_gemm` @ `topk_weights`
- `final_output`: if `topk > 1`, `topk_weight_mul.view(num_tokens, topk, -1).sum(dim=1)` else `topk_weight_mul`
One way to eliminate the loop is to use a grouped GEMM, where all expert GEMMs are computed within a single kernel, which iterates over tiles of the expert GEMMs as individual GEMMs, where each GEMM, the `A` matrix is `M' x K` and the `B` matrix is `K x N`, where `M'` is the number of tokens assigned to the expert and `B` is the weight matrix for that expert.
This requires an additional permute (and subsequent copy) of the hidden states such that the tokens assigned to each expert are contiguous in memory before running the first grouped GEMM within the Expert MLP.
Additionally, after the second grouped GEMM, the hidden states must be permuted back to the original token order and multiplied by `topk_weights` to get the final output.
### Optimizations
This repo implements a grouped GEMM-based MoE MLP with the following optimizations:
- Eliminates the loop over experts by performing gemms as a grouped GEMM, computing the expert gemms within a single fused triton kernel
- Fuses the permutation of hidden states from token order (original input order) to expert order (tokens grouped by expert) within the prologue of first the first grouped GEMM
- Fuses the (un)permutation of hidden states from expert order back to token order in second GEMM
- Fuses the mul of hidden states by expert weights within epilogue of second GEMM (only implemented for inference, not for training)
### Structure
- `grouped_gemm/interface.py`: wrappers for the individual forward / backward kernels as well as the `torch.autograd.Function`
- `grouped_gemm/kernels/forward.py`: forward kernel
- `grouped_gemm/kernels/backward.py`: backward dX and dW kernels
- `grouped_gemm/kernels/tuning.py`: manual tuning utils
- `grouped_gemm/kernels/autotuning.py`: autotuning utils
- `grouped_gemm/reference/moe_block.py`: contains `Qwen3MoeFusedGroupedGEMMBlock`, a reference implementation of Huggingface `Qwen3SparseMOEBlock` with fused triton kernel in-place of original HF expert computation
- `grouped_gemm/reference/moe_ops.py`: supporting ops (routing, token sorting, etc.) and reference MoE block using a torch-native grouped gemm approach.
### Tests
- `grouped_gemm/tests/test_grouped_gemm.py`: unit tests for forward, backward grouped gemm kernels as well as the wrapped grouped gemm autograd.Function. Best not to run this entire test suite at once due to the large number of parametrized unit tests. Rather, use filters to run specific
sets of tests. E.g., to run forward tests with autotune turned on: `pytest -sv -k "forward and autotune" --tb=short tests/test_grouped_gemm.py`. Use the test function names and parameter ids for words to filter on.
- `grouped_gemm/tests/test_qwen3_moe.py`: end to end test for Qwen3 MoE block. IMPORTANT: read `tests/run_qwen3_moe_tests.sh` as well as notes in the test itself for complications when running parametrized pytest test suites and triton / autotune. TLDR: use the test script and NOT pytest to run the tests.
### Benchmarks
- `grouped_gemm/benchmark/benchmark_fused_moe.py`: benchmarks HF `Qwen3SpareMOEBlock` against the fused implementation
Running with these flags on an `H100` to bench forward pass (run with `--help` to see all available flags):
```
python benchmark/benchmark_fused_moe.py --mode forward --seqlen 1024 --permute_x --permute_y --autotune
```
For the backward bench:
```
python benchmark/benchmark_fused_moe.py --mode backward --seqlen 1024 --permute_x --permute_y --autotune
```
On my machine and env, I get speedups > 25x and 14x respectively.
### Notes
- Tested and benched on `H100`, though should run on Ampere and possibly even earlier gpu generations though the autotuning configs will need to be adjusted.
- The env I used to develop the kernel was `pytorch 2.7/2.8` and `pytorch-triton 3.3`.
- The kernels can be run either as autotuned (see `autotuning.py`) or with manually specified config (see `tuning.py`). Recommended to run using autotuner since the MoE block requires 2 configs for the forward (2 grouped gemms) and 4 for the backwards (dX and dW per grouped gemm, 2 grouped gemms).
- Running with autotuning turned off with the default manual kernel config will result is **highly** sub-optimal performance as it is only meant for testing / debugging purposes.
- I've tried to strike a balance between compilation time and autotuning search space -- can probably squeeze even more performance for specific workloads.
TODO:
- TMA store: implemented but not enabled currently due to non-determinism arising from triton pipelining bug.
- Warp specialization: Hopper support for WS not yet enabled on triton 3.3x branch which ships with latest pytorch 2.7.
- Additional optimizations:
- Fused / optimized implementations of routing, token sorting, etc.
- Better software pipelining within grouped gemm
- Threadblock swizzling for better L2 caching

View file

View file

@ -0,0 +1,297 @@
import argparse
import time
import torch
from grouped_gemm.kernels.autotuning import (
DEFAULT_K_BLOCK_SIZES,
DEFAULT_M_BLOCK_SIZES,
DEFAULT_N_BLOCK_SIZES,
DEFAULT_NUM_STAGES,
DEFAULT_NUM_WARPS,
)
from grouped_gemm.kernels.tuning import (
KernelConfigBackward_dW,
KernelConfigBackward_dX,
KernelConfigForward,
KernelResult,
TritonTuningContext,
)
from grouped_gemm.reference.moe_block import Qwen3MoeFusedGroupedGEMMBlock
from transformers import AutoConfig
from transformers.models.qwen3_moe import Qwen3MoeConfig
from transformers.models.qwen3_moe.modeling_qwen3_moe import Qwen3MoeSparseMoeBlock
from triton.testing import do_bench
from utils import (
create_kernel_configs,
post_process_results,
save_results,
)
SEED = 42
def run_benchmark_forward(
config: Qwen3MoeConfig,
seqlen: int,
dtype: torch.dtype,
permute_x: bool,
permute_y: bool,
autotune: bool,
kernel_config_fwd: KernelConfigForward = None,
kernel_config_bwd_dW: KernelConfigBackward_dW = None,
kernel_config_bwd_dX: KernelConfigBackward_dX = None,
):
torch.manual_seed(SEED) # Should not be needed when running using pytest -- autouse fixture in conftest.py
device = "cuda"
hidden_size = config.hidden_size
bs = 1
# Reference op -- HF
moe_block = Qwen3MoeSparseMoeBlock(config).to(device, dtype)
# Triton kernel grouped gemm version of MoE Block -- this is what we're testing
fused_gemm_block = Qwen3MoeFusedGroupedGEMMBlock.from_hf(
moe_block,
permute_x=permute_x,
permute_y=permute_y,
autotune=autotune,
kernel_config_fwd=kernel_config_fwd,
kernel_config_bwd_dW=kernel_config_bwd_dW,
kernel_config_bwd_dX=kernel_config_bwd_dX,
).to(device, dtype)
X = torch.randn(bs, seqlen, hidden_size, dtype=dtype, device=device, requires_grad=True)
ref_output, _ = moe_block(X)
# Forward
bench_forward_ref = lambda: moe_block(X)
bench_forward_fused = lambda: fused_gemm_block(X)
ref_forward_time = do_bench(bench_forward_ref)
with TritonTuningContext(kernel_config_fwd) as ctx:
fused_forward_time = do_bench(bench_forward_fused)
if not ctx.success:
return 0, 1
print(
f"Forward: ref {ref_forward_time:.4f}, fused {fused_forward_time:.4f}, speedup {ref_forward_time / fused_forward_time:.1f}x"
)
return ref_forward_time, fused_forward_time
def run_benchmark_backward(
config: Qwen3MoeConfig,
seqlen: int,
dtype: torch.dtype,
permute_x: bool,
permute_y: bool,
autotune: bool,
kernel_config_fwd: KernelConfigForward = None,
kernel_config_bwd_dW: KernelConfigBackward_dW = None,
kernel_config_bwd_dX: KernelConfigBackward_dX = None,
dX_only: bool = False,
dW_only: bool = False,
):
torch.manual_seed(SEED) # Should not be needed when running using pytest -- autouse fixture in conftest.py
device = "cuda"
hidden_size = config.hidden_size
bs = 1
# Reference op -- HF
moe_block = Qwen3MoeSparseMoeBlock(config).to(device, dtype)
# Triton kernel grouped gemm version of MoE Block -- this is what we're testing
fused_gemm_block = Qwen3MoeFusedGroupedGEMMBlock.from_hf(
moe_block,
permute_x=permute_x,
permute_y=permute_y,
autotune=autotune,
kernel_config_fwd=kernel_config_fwd,
kernel_config_bwd_dW=kernel_config_bwd_dW,
kernel_config_bwd_dX=kernel_config_bwd_dX,
dX_only=dX_only,
dW_only=dW_only,
).to(device, dtype)
X = torch.randn(bs, seqlen, hidden_size, dtype=dtype, device=device, requires_grad=True)
X_test = X.detach().clone().requires_grad_(True)
output, _ = moe_block(X)
# Prevent autotuning forward pass
from grouped_gemm.kernels.forward import _autotuned_grouped_gemm_forward_kernel
_autotuned_grouped_gemm_forward_kernel.configs = _autotuned_grouped_gemm_forward_kernel.configs[:20]
test_output, _ = fused_gemm_block(X_test)
# Bench
grad_output = torch.randn_like(output)
bench_backward_ref = lambda: output.backward(grad_output, retain_graph=True) # noqa: E731
bench_backward_fused = lambda: test_output.backward(grad_output, retain_graph=True) # noqa: E731
ref_backward_time = do_bench(bench_backward_ref, grad_to_none=[X, *moe_block.parameters()])
fused_backward_time = do_bench(bench_backward_fused, grad_to_none=[X_test, *fused_gemm_block.parameters()])
print(
f"Backward: ref {ref_backward_time:.4f}, fused {fused_backward_time:.4f}, speedup {ref_backward_time / fused_backward_time:.1f}x"
)
return ref_backward_time, fused_backward_time
def run_benchmark(
mode: str,
model_config: Qwen3MoeConfig,
seqlen: int,
dtype: torch.dtype,
permute_x: bool,
permute_y: bool,
autotune: bool,
kernel_config_fwd: KernelConfigForward = None,
kernel_config_bwd_dW: KernelConfigBackward_dW = None,
kernel_config_bwd_dX: KernelConfigBackward_dX = None,
):
if mode == "forward":
ref_time, fused_time = run_benchmark_forward(
model_config,
seqlen,
dtype,
permute_x,
permute_y,
autotune,
kernel_config_fwd,
kernel_config_bwd_dW,
kernel_config_bwd_dX,
)
elif mode == "dW":
ref_time, fused_time = run_benchmark_backward(
model_config,
seqlen,
dtype,
permute_x,
permute_y,
autotune,
kernel_config_fwd,
kernel_config_bwd_dW,
kernel_config_bwd_dX,
dW_only=True,
)
elif mode == "dX":
ref_time, fused_time = run_benchmark_backward(
model_config,
seqlen,
dtype,
permute_x,
permute_y,
autotune,
kernel_config_fwd,
kernel_config_bwd_dW,
kernel_config_bwd_dX,
dX_only=True,
)
elif mode == "backward":
ref_time, fused_time = run_benchmark_backward(
model_config,
seqlen,
dtype,
permute_x,
permute_y,
autotune,
kernel_config_fwd,
kernel_config_bwd_dW,
kernel_config_bwd_dX,
dX_only=False,
dW_only=False,
)
return ref_time, fused_time
# NOTE: better to use autotuner for now, since the MoE block needs 2 different kernel configs for forward (2 grouped gemms, gate_up_proj and down_proj)
# and the backward pass needs 4 different kernel configs (2 grouped gemms each for dW and dX)
# The benchmark only supports 1 kernel config at a time so the same config will be used for both grouped gemms, which is suboptimal.
if __name__ == "__main__":
parser = argparse.ArgumentParser()
parser.add_argument("--results_dir", type=str, default="benchmark_results")
parser.add_argument("--seqlen", type=int, default=1024)
parser.add_argument("--dtype", type=str, choices=["bfloat16", "float16"], default="bfloat16")
parser.add_argument("--permute_x", action="store_true")
parser.add_argument("--permute_y", action="store_true")
parser.add_argument("--autotune", action="store_true")
parser.add_argument("--BLOCK_SIZE_M", nargs=2, type=int, default=[DEFAULT_M_BLOCK_SIZES[0], DEFAULT_M_BLOCK_SIZES[-1]])
parser.add_argument("--BLOCK_SIZE_N", nargs=2, type=int, default=[DEFAULT_N_BLOCK_SIZES[0], DEFAULT_N_BLOCK_SIZES[-1]])
parser.add_argument("--BLOCK_SIZE_K", nargs=2, type=int, default=[DEFAULT_K_BLOCK_SIZES[0], DEFAULT_K_BLOCK_SIZES[-1]])
parser.add_argument("--num_warps", nargs=2, type=int, default=[DEFAULT_NUM_WARPS[0], DEFAULT_NUM_WARPS[-1]])
parser.add_argument("--num_stages", nargs=2, type=int, default=[DEFAULT_NUM_STAGES[0], DEFAULT_NUM_STAGES[-1]])
parser.add_argument("--use_tma_load_w", action="store_true") # No need to specify, will automatically parametrize these for each kernel config
parser.add_argument("--use_tma_load_x", action="store_true") # No need to specify, will automatically parametrize these for each kernel config
parser.add_argument("--use_tma_load_dy", action="store_true") # No need to specify, will automatically parametrize these for each kernel config
parser.add_argument("--mode", type=str, choices=["forward", "backward", "dW", "dX"], default="forward")
args = parser.parse_args()
args.dtype = getattr(torch, args.dtype)
model_id = "Qwen/Qwen3-30B-A3B"
model_config = AutoConfig.from_pretrained(model_id)
mode = args.mode
if args.autotune:
print(
f"Benchmarking {model_id} {mode}: seqlen={args.seqlen}, dtype={args.dtype}, permute_x={args.permute_x}, permute_y={args.permute_y}, autotune"
)
start_time = time.time()
ref_time, fused_time = run_benchmark(
args.mode,
model_config,
seqlen=args.seqlen,
dtype=args.dtype,
permute_x=args.permute_x,
permute_y=args.permute_y,
autotune=args.autotune,
)
end_time = time.time()
print(f"Total time: {end_time - start_time:.4f} seconds")
else:
kernel_configs = create_kernel_configs(args, args.permute_x, args.permute_y)
print(f"Running {len(kernel_configs)} kernel configs")
default_kernel_config_fwd = KernelConfigForward(permute_x=args.permute_x, permute_y=args.permute_y)
default_kernel_config_bwd_dW = KernelConfigBackward_dW(permute_x=args.permute_x, permute_y=args.permute_y)
default_kernel_config_bwd_dX = KernelConfigBackward_dX(permute_x=args.permute_x, permute_y=args.permute_y)
results = []
for kernel_config in kernel_configs:
if args.mode == "forward":
kernel_config_fwd = kernel_config
kernel_config_bwd_dW = default_kernel_config_bwd_dW
kernel_config_bwd_dX = default_kernel_config_bwd_dX
elif args.mode == "dW":
kernel_config_fwd = default_kernel_config_fwd
kernel_config_bwd_dW = kernel_config
kernel_config_bwd_dX = default_kernel_config_bwd_dX
elif args.mode == "dX":
kernel_config_fwd = default_kernel_config_fwd
kernel_config_bwd_dW = default_kernel_config_bwd_dW
kernel_config_bwd_dX = kernel_config
else:
raise ValueError(f"Invalid mode: {args.mode}")
print(
f"Benchmarking {model_id} {args.mode} with seqlen={args.seqlen}, dtype={args.dtype}, permute_x={args.permute_x}, permute_y={args.permute_y}, kernel_config_fwd={kernel_config_fwd}, kernel_config_bwd_dW={kernel_config_bwd_dW}, kernel_config_bwd_dX={kernel_config_bwd_dX}"
)
ref_time, fused_time = run_benchmark(
args.mode,
model_config,
seqlen=args.seqlen,
dtype=args.dtype,
permute_x=kernel_config.permute_x,
permute_y=kernel_config.permute_y,
autotune=False,
kernel_config_fwd=kernel_config_fwd,
kernel_config_bwd_dW=kernel_config_bwd_dW,
kernel_config_bwd_dX=kernel_config_bwd_dX,
)
results.append(KernelResult(
torch_time=ref_time,
triton_time=fused_time,
speedup=ref_time / fused_time,
kernel_config=kernel_config,
))
df = post_process_results(results, args.mode, args.seqlen, args.dtype, args.autotune)
save_results(df, args.results_dir, args.mode, args.seqlen, args.dtype, args.autotune)

View file

@ -0,0 +1,143 @@
import argparse
import datetime
import json
import logging
import math
import os
from itertools import product
import pandas as pd
import torch
from grouped_gemm.kernels.tuning import (
KernelConfigBackward_dW,
KernelConfigBackward_dX,
KernelConfigForward,
KernelResult,
)
SEED = 42
def create_merged_results(
df: pd.DataFrame, mode: str, seqlen: int, dtype: torch.dtype, autotune: bool
):
kernel_result_cols = df.columns.to_list()
test_config_dict = {"mode": mode, "seqlen": seqlen, "dtype": dtype, "autotune": autotune}
test_config_cols = list(test_config_dict.keys())
for col in test_config_cols:
df[col] = test_config_dict[col]
# Reorder columns so that test config cols are first
df = df[test_config_cols + kernel_result_cols]
return df
def post_process_results(
results: list[KernelResult],
mode: str,
seqlen: int,
dtype: torch.dtype,
autotune: bool,
):
df = KernelResult.to_dataframe(results, sort_by="speedup")
df = create_merged_results(df, mode, seqlen, dtype, autotune)
return df
def save_results(
df: pd.DataFrame,
results_dir: str,
mode: str,
seqlen: int,
dtype: torch.dtype,
autotune: bool,
):
dt = datetime.datetime.now().strftime("%Y%m%d_%H%M")
save_dir = f"{results_dir}/{mode}"
save_path = f"{save_dir}/{dt}_{seqlen}_{str(dtype).split('.')[-1]}.csv"
if not os.path.exists(save_dir):
os.makedirs(save_dir)
print(f"Saving results to {save_path}")
df.to_csv(save_path, index=False)
def create_kernel_configs(args: argparse.Namespace, permute_x: bool, permute_y: bool):
block_m_range = power_of_two_range(args.BLOCK_SIZE_M[0], args.BLOCK_SIZE_M[1])
block_n_range = power_of_two_range(args.BLOCK_SIZE_N[0], args.BLOCK_SIZE_N[1])
block_k_range = power_of_two_range(args.BLOCK_SIZE_K[0], args.BLOCK_SIZE_K[1])
num_warps_range = multiples_of_range(args.num_warps[0], args.num_warps[1], step=2)
num_stages_range = multiples_of_range(args.num_stages[0], args.num_stages[1], step=1)
mode = args.mode
kernel_configs = []
for block_m, block_n, block_k, num_warps, num_stages, tma_load_a, tma_load_b in product(
block_m_range, block_n_range, block_k_range, num_warps_range, num_stages_range, [True, False], [True, False]
):
if mode == "forward":
kernel_config = KernelConfigForward(
BLOCK_SIZE_M=block_m,
BLOCK_SIZE_N=block_n,
BLOCK_SIZE_K=block_k,
num_warps=num_warps,
num_stages=num_stages,
use_tma_load_w=tma_load_a,
use_tma_load_x=tma_load_b,
permute_x=permute_x,
permute_y=permute_y,
)
elif mode == "dW":
kernel_config = KernelConfigBackward_dW(
BLOCK_SIZE_M=block_m,
BLOCK_SIZE_N=block_n,
BLOCK_SIZE_K=block_k,
num_warps=num_warps,
num_stages=num_stages,
use_tma_load_dy=tma_load_a,
use_tma_load_x=tma_load_b,
permute_x=permute_x,
permute_y=permute_y,
)
elif mode == "dX":
kernel_config = KernelConfigBackward_dX(
BLOCK_SIZE_M=block_m,
BLOCK_SIZE_N=block_n,
BLOCK_SIZE_K=block_k,
num_warps=num_warps,
num_stages=num_stages,
use_tma_load_dy=tma_load_a,
use_tma_load_w=tma_load_b,
permute_x=permute_x,
permute_y=permute_y,
)
else:
raise ValueError(f"Invalid mode: {mode}")
kernel_configs.append(kernel_config)
logging.info(f"Pruning {len(kernel_configs)} kernel configs")
pruned_configs = []
for config in kernel_configs:
if mode == "forward":
if permute_x and config.use_tma_load_x:
continue
elif mode == "dW":
if permute_x and config.use_tma_load_x:
continue
if permute_y and config.use_tma_load_dy:
continue
elif mode == "dX":
if permute_y and config.use_tma_load_dy:
continue
pruned_configs.append(config)
logging.info(f"After pruning, {len(pruned_configs)} kernel configs")
return pruned_configs
def power_of_two_range(start, end):
start = math.log2(start)
end = math.log2(end)
return [2**i for i in range(int(start), int(end) + 1)]
def multiples_of_range(start, end, step=1):
return list(range(start, end + step, step))

View file

@ -0,0 +1,965 @@
import logging
import warnings
from dataclasses import asdict
import torch
import triton
from grouped_gemm.kernels.backward import (
_autotuned_grouped_gemm_dW_kernel,
_autotuned_grouped_gemm_dX_kernel,
_grouped_gemm_dW_kernel,
_grouped_gemm_dX_kernel,
)
from grouped_gemm.kernels.forward import (
_autotuned_grouped_gemm_forward_kernel,
_grouped_gemm_forward_kernel,
)
from grouped_gemm.kernels.tuning import (
KernelConfigBackward_dW,
KernelConfigBackward_dX,
KernelConfigForward,
)
logger = logging.getLogger(__name__)
# Set formatter to include timestamp, pathname and lineno
formatter = logging.Formatter(
"%(asctime)s::%(levelname)s,%(pathname)s:%(lineno)d:: %(message)s"
)
# Add console handler
ch = logging.StreamHandler()
ch.setFormatter(formatter)
logger.addHandler(ch)
_FUSED_MUL_WARN = False
_SUPPORTS_TMA = None
def supports_tma():
global _SUPPORTS_TMA
if _SUPPORTS_TMA is None:
_SUPPORTS_TMA = torch.cuda.get_device_capability()[0] >= 9
return _SUPPORTS_TMA
_per_device_alloc_fns = {}
def get_per_device_per_stream_alloc_fn(device):
if device not in _per_device_alloc_fns:
_per_stream_tensors = {}
def alloc_fn(size: int, alignment: int, stream):
assert alignment == 128
if (
stream not in _per_stream_tensors
or _per_stream_tensors[stream].numel() < size
):
_per_stream_tensors[stream] = torch.empty(
size, device=device, dtype=torch.int8
)
_per_stream_tensors[stream].__hibernate__ = {"type": "ignore"}
return _per_stream_tensors[stream]
_per_device_alloc_fns[device] = alloc_fn
return _per_device_alloc_fns[device]
def log_kernel_info(
compiled_kernel: triton.compiler.CompiledKernel, best_config: triton.Config = None
):
kernel_name = compiled_kernel.name
nregs = compiled_kernel.n_regs
nspills = compiled_kernel.n_spills
metadata = compiled_kernel.metadata
logger.debug(
f"{kernel_name}: n_regs={nregs} n_spills={nspills} metadata={metadata}"
)
if best_config is not None:
logger.debug(f"{kernel_name} autotuned best_config: {best_config}")
def grouped_gemm_forward(
X: torch.Tensor,
W: torch.Tensor,
topk: int,
m_sizes: torch.Tensor,
gather_indices: torch.Tensor = None,
topk_weights: torch.Tensor = None,
# Fusions
permute_x: bool = False,
permute_y: bool = False,
fuse_mul_post: bool = False,
# Autotuning - manual kernel params will be ignored if autotune is True
autotune: bool = False,
# Kernel tuning params if not autotuning -- NOTE: these params need to be tuned, otherwise performance will be poor
BLOCK_SIZE_M: int = 32,
BLOCK_SIZE_N: int = 32,
BLOCK_SIZE_K: int = 32,
num_warps: int = 4,
num_stages: int = 2,
use_tma_load_w: bool = False,
use_tma_load_x: bool = False,
use_tma_store: bool = False,
# software pipelining -- set to True for now, won't impact until loop is re-written
flatten: bool = True,
# debugging
debug: bool = False,
) -> torch.Tensor:
"""
Grouped GEMM forward pass for MoE MLPs.
The implementation offers a number of fusions specific to MoE:
- `permute_x`: fuse the permutation of hidden states from token order (original order) to grouped expert order, typically only needed for the first grouped GEMM in an MoE MLP.
- When `permute_x` is True, `X` is expected to be of shape (num_tokens, K).
- When `permute_x` is False, `X` is expected to be of shape (total_tokens, K) where `total_tokens = num_tokens * topk` AND already permuted to grouped expert order, i.e., hidden states are sorted such that tokens assigned to each expert are contiguous.
- `permute_y`: fused the permuation of the output from expert grouped order back to original token order, typically only needed for the second grouped GEMM in an MoE MLP.
- `fuse_mul_pre`: fuse the multiplication of the routed input with topk_weights, only done in the first grouped GEMM in an MoE MLP as for Llama4. Do not use, since results in performance regression as it interrupts the GEMM mainloop.
- `fuse_mul_post`: fuse the multiplication of the routed output with topk_weights, used only when `permute_y` is True. NOTE: this should only be used when using this kernel for inference, not for training.
X: (M, K) hidden states where M is the num_tokens if `permute_x` is True, otherwise `total_tokens` where `total_tokens = num_tokens * topk`.
W: (E, N, K) expert weights, where E is number of experts, N in the intermediate (output) dim, and K is the reduction dim
m_sizes: tokens assigned to each expert which correspond to the size of M in the respective GEMMs in the grouped GEMM.
gather_indices: (total_tokens,) indices of tokens assigned to each expert. E.g., slicing gather_indices by cumsum of m_sizes gives the indices of tokens assigned to each expert.
topk_weights: (total_tokens,) weights to multiply routed output by in expert MLP calculation, used only when `fuse_mul_post` is True (see note on `fuse_mul_post`).
use_fast_accum: currently unused; trade off faster accumulation dtype in GEMM for less precision.
use_tma_load_x: use TMA for loading activations, incompatible with permute_x. TODO: add TMA gather / scatter support for Blackwell+.
use_tma_load_w: use TMA for loading weights. If TMA supported, this should always be enabled as it is faster than global memory load.
use_tma_store: use TMA for storing output, incompatible with permute_y. TODO: add TMA scatter support for Blackwell+.
Returns:
y: (total_tokens, N) output of grouped GEMM
"""
assert X.device.type == "cuda", "X and W must be on CUDA"
assert m_sizes.device.type == "cuda", "m_sizes must be on CUDA"
X = X.contiguous()
W = W.contiguous()
m_sizes = m_sizes.contiguous()
# Preconditions
assert not (permute_x and permute_y), "Cannot permute both X and Y"
assert not (permute_y and use_tma_store), "Cannot use both TMA store and permute_y"
if use_tma_load_x:
# TMA load for activations, TMA gather only supported on Blackwell+
assert not permute_x, "Cannot use both use_tma_load_x and permute_x"
use_tma = use_tma_load_w or use_tma_load_x or use_tma_store
if not supports_tma() and use_tma:
warnings.warn("TMA not supported, tma_load will be set to False")
use_tma_load_w = False
use_tma_load_x = False
use_tma_store = False
if use_tma or autotune:
def alloc_fn(size: int, alignment: int, stream: int):
return torch.empty(size, device="cuda", dtype=torch.int8)
triton.set_allocator(alloc_fn)
X = X.view(-1, X.shape[-1])
W = W.view(-1, W.shape[-1])
if permute_x or permute_y:
assert gather_indices is not None, (
"gather_indices must be provided when permute_x or permute_y is True"
)
assert gather_indices.is_contiguous()
assert gather_indices.device.type == "cuda"
assert gather_indices.ndim == 1
total_tokens = gather_indices.shape[0]
num_tokens = total_tokens // topk
if permute_x:
assert X.shape[0] == num_tokens, (
f"X.shape[0] ({X.shape[0]}) must match num_tokens ({num_tokens})"
)
else:
assert X.shape[0] == total_tokens, (
f"X.shape[0] ({X.shape[0]}) must match total_tokens ({total_tokens})"
)
else:
total_tokens = X.shape[0]
num_tokens = total_tokens // topk
num_experts = m_sizes.shape[0]
_, K = X.shape
N = W.shape[0] // num_experts
assert K == W.shape[1], f"K ({K}) must match W.shape[1] ({W.shape[1]})"
if fuse_mul_post:
global _FUSED_MUL_WARN
if not _FUSED_MUL_WARN:
warnings.warn(
"fused_mul should only be used for inference, not for training"
)
_FUSED_MUL_WARN = True
assert permute_y, "FUSE_MUL requires PERMUTE_Y"
assert topk_weights is not None
assert topk_weights.numel() == total_tokens
assert topk_weights.device.type == "cuda"
assert topk_weights.is_contiguous()
topk_weights = topk_weights.view(-1)
if debug:
print(
f"DEBUG::GROUPED_GEMM {topk_weights.tolist()} {gather_indices.tolist()}"
)
y = torch.empty((total_tokens, N), device=X.device, dtype=X.dtype)
if total_tokens == 0 or N == 0:
return y
NUM_SMS = torch.cuda.get_device_properties("cuda").multi_processor_count
def grid(META):
return (NUM_SMS,)
if not autotune:
BLOCK_SIZE_K = min(K, BLOCK_SIZE_K)
BLOCK_SIZE_N = min(N, BLOCK_SIZE_N)
BLOCK_SIZE_M = min(total_tokens, BLOCK_SIZE_M)
if debug:
print(
f"DEBUG::GROUPED_GEMM {num_tokens=} {topk=} {num_experts=} {N=} {K=} {BLOCK_SIZE_M=} {BLOCK_SIZE_N=} {BLOCK_SIZE_K=} {permute_x=}"
)
print(
f"DEBUG::GROUPED_GEMM {m_sizes.tolist()} {(gather_indices // topk).tolist()}"
)
kernel_args = {
# Inputs
"x_ptr": X,
"w_ptr": W,
"m_sizes_ptr": m_sizes,
"gather_indices_ptr": gather_indices,
"topk_weights_ptr": topk_weights,
# Output
"y_ptr": y,
# Problem shapes
"NUM_TOKENS": num_tokens,
"NUM_EXPERTS": num_experts,
"TOPK": topk,
"N": N,
"K": K,
"NUM_SMS": NUM_SMS,
# Gather / Scatter
"PERMUTE_X": permute_x,
"PERMUTE_Y": permute_y,
# TopK weight merging
"FUSE_MUL_POST": fuse_mul_post,
# Loop pipelining
"FLATTEN": flatten,
}
if not autotune:
kernel_args.update(
{
"USE_TMA_LOAD_W": use_tma_load_w,
"USE_TMA_LOAD_X": use_tma_load_x,
"USE_TMA_STORE": use_tma_store,
"BLOCK_SIZE_M": BLOCK_SIZE_M,
"BLOCK_SIZE_N": BLOCK_SIZE_N,
"BLOCK_SIZE_K": BLOCK_SIZE_K,
"num_warps": num_warps,
"num_stages": num_stages,
}
)
kernel = (
_autotuned_grouped_gemm_forward_kernel
if autotune
else _grouped_gemm_forward_kernel
)
compiled_kernel: triton.compiler.CompiledKernel = kernel[grid](**kernel_args)
if autotune:
log_kernel_info(compiled_kernel, kernel.best_config)
else:
log_kernel_info(compiled_kernel)
return y
def grouped_gemm_dX(
dY: torch.Tensor,
W: torch.Tensor,
gather_indices: torch.Tensor,
m_sizes: torch.Tensor,
topk: int,
BLOCK_SIZE_M: int = 32,
BLOCK_SIZE_N: int = 32,
BLOCK_SIZE_K: int = 32,
debug: bool = False,
permute_x: bool = False,
permute_y: bool = False,
use_tma_load_w: bool = False,
use_tma_load_dy: bool = False,
use_tma_store: bool = False,
num_warps: int = 4,
num_stages: int = 2,
flatten: bool = True,
fuse_mul_pre: bool = False,
fuse_mul_post: bool = False,
autotune: bool = False,
) -> torch.Tensor:
"""
dX backward kernel
grad_output: (M, N)
gather_indices: (total_tokens,), indices of tokens assigned to each expert. E.g., slicing gather_indices by cumsum of m_sizes gives the indices of tokens assigned to each expert.
m_sizes: tokens assigned to each expert which correspond to the size of M in the respective GEMMs in the grouped GEMM.
topk: number of experts chosen per token.
`permute_x`: whether X was permuted on load in the forward pass, typically only used for the first grouped GEMM in an MoE MLP to group tokens by expert.
- In the forward pass, if we permuted X on load, we need to permute store in the backward pass
- Shapes
- the forward pass input X shape is [NUM_TOKENS, K], reduce across K, output y is [NUM_TOKENS * TOPK, K]
- the backward pass input dy shape is [NUM_TOKENS * TOPK, N], reduce across N, output dX is [NUM_TOKENS * TOPK, K]
- Note that in the backward pass, the output size is still [NUM_TOKENS * TOPK, K] since we still need to accumulate gradients for each expert chosen by the token in a post-processing step.
`permute_y`: whether the output was permuted on store in the forward pass, typically only used for the second grouped GEMM in an MoE MLP to restore to the original token order.
- In the forward pass, if we permuted output on store (e.g., in the second grouped GEMM in fused MoE MLP), we need to permute on load to get from token order to expert grouped order
- We still store in contiguous order since we are writing out dX which will be the input to the backwards pass of the first grouped GEMM
`fuse_mul_{pre,post}`: always set to False since this should only be used for inference.
use_tma_load_dy: use TMA for loading dy. use_tma_load_dy is incompatible with permute_y. TODO: add TMA gather / scatter support for Blackwell+ which will enable permute_y and use_tma_load_dy.
use_tma_load_w: use TMA for loading weights. If TMA supported, this should always be enabled as it is faster than global memory load.
use_tma_store: use TMA for storing dX. Incompatible with permute_x. TODO: add TMA gather / scatter support for Blackwell+ which will enable permute_x and use_tma_store.
"""
assert not fuse_mul_pre, (
"fuse_mul_pre should only be used for inference, not for training"
)
assert not fuse_mul_post, (
"fuse_mul_post should only be used for inference, not for training"
)
assert dY.is_contiguous()
assert W.is_contiguous()
assert m_sizes.is_contiguous()
assert m_sizes.ndim == 1
# Preconditions
assert not (permute_x and permute_y), "Cannot permute both X and Y"
# Note that this is flipped from the forward pass
# If we permuted y in the forward, we need to permute on load in the backward
assert not (permute_y and use_tma_load_dy), "Cannot use both TMA load and permute_y"
assert not (permute_x and use_tma_store), "Cannot use both TMA store and permute_x"
use_tma = use_tma_load_dy or use_tma_load_w or use_tma_store
if not supports_tma() and use_tma:
warnings.warn("TMA not supported, tma_load will be set to False")
use_tma_load_w = False
use_tma_load_dy = False
use_tma_store = False
if use_tma or autotune:
def alloc_fn(size: int, alignment: int, stream: int):
# print(f"DEBUG::GROUPED_GEMM alloc_fn {size=} {alignment=} {stream=}")
return torch.empty(size, device="cuda", dtype=torch.int8)
triton.set_allocator(alloc_fn)
num_experts = m_sizes.shape[0]
dY = dY.view(-1, dY.shape[-1])
W = W.view(-1, W.shape[-1])
M_total, N_grad = dY.shape
N_total, K = W.shape
N = N_total // num_experts
assert N_grad == N, f"Grad_output N ({N_grad}) must match weight N ({N})"
assert M_total % topk == 0, (
f"M_total ({M_total}) must be divisible by topk ({topk})"
)
num_tokens = M_total // topk
total_tokens = gather_indices.shape[0]
assert total_tokens == M_total, (
f"Total tokens ({total_tokens}) must match M_total ({M_total})"
)
# Note that the output shape is [NUM_TOKENS * TOPK, K] even when `permute_x` is True since we need to accumulate gradients across all experts chosen by the token.
# This will be done in a post-processing step reduction step.
output_shape = (total_tokens, K)
dX = torch.zeros(output_shape, device=dY.device, dtype=dY.dtype)
NUM_SMS = torch.cuda.get_device_properties(
"cuda"
).multi_processor_count # if not debug else 1
def grid(META):
return (NUM_SMS,)
if not autotune:
BLOCK_SIZE_M = min(M_total, BLOCK_SIZE_M)
BLOCK_SIZE_N = min(N_grad, BLOCK_SIZE_N)
BLOCK_SIZE_K = min(K, BLOCK_SIZE_K)
if debug:
print(
f"DEBUG::GROUPED_GEMM {num_tokens=} {topk=} {output_shape=} {num_experts=} {N=} {K=} {BLOCK_SIZE_M=} {BLOCK_SIZE_N=} {BLOCK_SIZE_K=} {NUM_SMS=}"
)
print(f"DEBUG::GROUPED_GEMM {m_sizes.tolist()}")
kernel_args = {
# Inputs
"dY_ptr": dY,
"w_ptr": W,
"gather_indices_ptr": gather_indices,
"m_sizes_ptr": m_sizes,
# Output
"dX_ptr": dX,
# Problem sizes
"NUM_EXPERTS": num_experts,
"NUM_TOKENS": num_tokens,
"TOPK": topk,
"N": N,
"K": K,
"NUM_SMS": NUM_SMS,
# Gather / Scatter
"PERMUTE_X": permute_x,
"PERMUTE_Y": permute_y,
"FLATTEN": flatten,
}
if not autotune:
kernel_args.update(
{
"BLOCK_SIZE_M": BLOCK_SIZE_M,
"BLOCK_SIZE_N": BLOCK_SIZE_N,
"BLOCK_SIZE_K": BLOCK_SIZE_K,
"num_warps": num_warps,
"num_stages": num_stages,
"USE_TMA_LOAD_dY": use_tma_load_dy,
"USE_TMA_LOAD_W": use_tma_load_w,
"USE_TMA_STORE": use_tma_store,
}
)
kernel = _autotuned_grouped_gemm_dX_kernel if autotune else _grouped_gemm_dX_kernel
compiled_kernel: triton.compiler.CompiledKernel = kernel[grid](**kernel_args)
if autotune:
log_kernel_info(compiled_kernel, kernel.best_config)
else:
log_kernel_info(compiled_kernel)
return dX
def grouped_gemm_dW(
X: torch.Tensor,
dY: torch.Tensor,
m_sizes: torch.Tensor,
gather_indices: torch.Tensor,
topk: int,
BLOCK_SIZE_M: int = 32,
BLOCK_SIZE_N: int = 32,
BLOCK_SIZE_K: int = 32,
permute_x: bool = False,
permute_y: bool = False,
use_tma_load_dy: bool = False,
use_tma_load_x: bool = False,
use_tma_store: bool = False,
fuse_mul_pre: bool = False,
fuse_mul_post: bool = False,
num_warps: int = 4,
num_stages: int = 2,
flatten: bool = True,
autotune: bool = False,
debug: bool = False,
) -> torch.Tensor:
"""
X: (M, K) hidden states where M is the num_tokens if `permute_x` is True, otherwise `total_tokens` where `total_tokens = num_tokens * topk`.
dY: (M, N)
topk: number of experts to choose per token.
m_sizes: tokens assigned to each expert which correspond to the size of M in the respective GEMMs in the grouped GEMM.
gather_indices: (total_tokens,) indices of tokens assigned to each expert. E.g., slicing gather_indices by cumsum of m_sizes gives the indices of tokens assigned to each expert.
permute_x: whether X was permuted on load in the forward pass, typically only used for the first grouped GEMM in an MoE MLP to group tokens by expert.
- for the first grouped GEMM, we permuted on load -> X was [num_tokens, K] and stored y in expert grouped order [num_tokens * topk, K]
- in the backwards pass, we need to permute on load of X while loading dy in contiguous (expert grouped) order
- since we are writing out dW, there is no need to permute on store
permute_y: whether the output was permuted on store in the forward pass, typically only used for the second grouped GEMM in an MoE MLP to restore to the original token order.
- for the second grouped GEMM, we permuted on store -> y was permuted from expert grouped order to token order while X was loaded in expert grouped order since it was the output of the first grouped GEMM
- in the backwards pass, we need to permute on load of dy to get from token order to expert grouped order to match the order of X
- since we are writing out dW, there is no need to permute on store
use_tma_load_dy: use TMA for loading dy. use_tma_load_dy is incompatible with permute_y. TODO: add TMA gather / scatter support for Blackwell+ which will enable permute_y and use_tma_load_dy.
use_tma_load_x: use TMA for loading x. use_tma_load_x is incompatible with permute_x. TODO: add TMA gather / scatter support for Blackwell+ which will enable permute_x and use_tma_load_x.
use_tma_store: use TMA for storing dW. If TMA supported, this should always be enabled as it is faster than global memory store.
"""
assert not fuse_mul_pre, "fuse_mul_pre not supported"
assert not fuse_mul_post, "fuse_mul_post not supported"
NUM_SMS = (
torch.cuda.get_device_properties("cuda").multi_processor_count
if not debug
else 1
)
X = X.view(-1, X.shape[-1]).contiguous()
dY = dY.contiguous()
m_sizes = m_sizes.contiguous()
# Preconditions
assert not (permute_x and permute_y), "Cannot permute both X and Y"
assert not (permute_y and use_tma_load_dy), "Cannot use both TMA load and permute_y"
assert not (permute_x and use_tma_load_x), "Cannot use both TMA load and permute_x"
use_tma = use_tma_load_dy or use_tma_load_x or use_tma_store
if not supports_tma() and use_tma:
warnings.warn("TMA not supported, tma_load will be set to False")
use_tma_load_x = False
use_tma_load_dy = False
use_tma_store = False
if use_tma or autotune:
def alloc_fn(size: int, alignment: int, stream: int):
return torch.empty(size, device="cuda", dtype=torch.int8)
triton.set_allocator(alloc_fn)
if permute_x or permute_y:
assert gather_indices is not None
assert gather_indices.is_contiguous()
assert gather_indices.device.type == "cuda"
assert gather_indices.ndim == 1
total_tokens = gather_indices.shape[0]
num_tokens = total_tokens // topk
if permute_x:
assert X.shape[0] == num_tokens
else:
assert X.shape[0] == total_tokens
else:
total_tokens = X.shape[0]
num_tokens = total_tokens // topk
num_experts = m_sizes.shape[0]
# Get dimensions
_, K = X.shape
M_grad, N = dY.shape
assert M_grad == total_tokens, f"dY M ({M_grad}) != total_tokens ({total_tokens})"
dW = torch.zeros((num_experts, N, K), device=X.device, dtype=X.dtype)
if not autotune:
BLOCK_SIZE_M = min(total_tokens, BLOCK_SIZE_M)
BLOCK_SIZE_N = min(N, BLOCK_SIZE_N)
BLOCK_SIZE_K = min(K, BLOCK_SIZE_K)
def grid(META):
return (NUM_SMS,)
if debug:
print(
f"DEBUG::GROUPED_GEMM_DW_TMA {num_experts=} {N=} {K=} {BLOCK_SIZE_M=} {BLOCK_SIZE_N=} {BLOCK_SIZE_K=} {NUM_SMS=}"
)
print(f"DEBUG::GROUPED_GEMM_DW_TMA {m_sizes.tolist()=}")
print(f"DEBUG::GROUPED_GEMM_DW_TMA {gather_indices.tolist()=}")
m_start = 0
for i in range(num_experts):
expert_token_idx = gather_indices[m_start : m_start + m_sizes[i]]
t_start = 0
while t_start < m_sizes[i]:
token_idx = expert_token_idx[t_start : t_start + BLOCK_SIZE_M]
if permute_x:
token_idx = token_idx // topk
print(
f"DEBUG::GROUPED_GEMM_DW_TMA Token expert {i} indices: {token_idx.tolist()}"
)
t_start += BLOCK_SIZE_M
m_start += m_sizes[i]
kernel_args = {
# Inputs
"x_ptr": X,
"dY_ptr": dY,
"m_sizes_ptr": m_sizes,
"gather_indices_ptr": gather_indices,
# Output
"dW_ptr": dW,
# Problem sizes
"NUM_TOKENS": num_tokens,
"TOPK": topk,
"NUM_EXPERTS": num_experts,
"N": N,
"K": K,
"NUM_SMS": NUM_SMS,
# Gather / Scatter
"PERMUTE_X": permute_x,
"PERMUTE_Y": permute_y,
# Loop pipelining
"FLATTEN": flatten,
}
if not autotune:
kernel_args.update(
{
"BLOCK_SIZE_M": BLOCK_SIZE_M,
"BLOCK_SIZE_N": BLOCK_SIZE_N,
"BLOCK_SIZE_K": BLOCK_SIZE_K,
"USE_TMA_LOAD_dY": use_tma_load_dy,
"USE_TMA_LOAD_X": use_tma_load_x,
"USE_TMA_STORE": use_tma_store,
"num_warps": num_warps,
"num_stages": num_stages,
}
)
kernel = _autotuned_grouped_gemm_dW_kernel if autotune else _grouped_gemm_dW_kernel
compiled_kernel: triton.compiler.CompiledKernel = kernel[grid](**kernel_args)
if autotune:
log_kernel_info(compiled_kernel, kernel.best_config)
else:
log_kernel_info(compiled_kernel)
return dW
class GroupedGemm(torch.autograd.Function):
@staticmethod
def forward(
ctx,
X,
W,
m_sizes,
topk,
gather_indices,
permute_x,
permute_y,
topk_weights,
fuse_mul_post,
kernel_config_fwd,
kernel_config_bwd_dX,
kernel_config_bwd_dW,
autotune,
dX_only,
dW_only,
):
ctx.topk = topk
ctx.permute_x = permute_x
ctx.permute_y = permute_y
ctx.fuse_mul_post = fuse_mul_post
ctx.kernel_config_fwd = kernel_config_fwd
ctx.kernel_config_bwd_dX = kernel_config_bwd_dX
ctx.kernel_config_bwd_dW = kernel_config_bwd_dW
ctx.autotune = autotune
ctx.dX_only = dX_only
ctx.dW_only = dW_only
# NOTE: we don't save topk_weights for backward since we do not support training with fused_mul
ctx.save_for_backward(X, W, m_sizes, gather_indices)
fwd_config = {}
if kernel_config_fwd is not None:
fwd_config["BLOCK_SIZE_M"] = kernel_config_fwd.BLOCK_SIZE_M
fwd_config["BLOCK_SIZE_N"] = kernel_config_fwd.BLOCK_SIZE_N
fwd_config["BLOCK_SIZE_K"] = kernel_config_fwd.BLOCK_SIZE_K
fwd_config["num_warps"] = kernel_config_fwd.num_warps
fwd_config["num_stages"] = kernel_config_fwd.num_stages
fwd_config["use_tma_load_x"] = kernel_config_fwd.use_tma_load_x
fwd_config["use_tma_load_w"] = kernel_config_fwd.use_tma_load_w
fwd_config["use_tma_store"] = kernel_config_fwd.use_tma_store
return grouped_gemm_forward(
X=X,
W=W,
topk=topk,
m_sizes=m_sizes,
gather_indices=gather_indices,
topk_weights=topk_weights,
permute_x=permute_x,
permute_y=permute_y,
fuse_mul_post=fuse_mul_post,
# Autotune -- this will override the manual kernel config if true
autotune=autotune,
# Manual kernel config
**fwd_config,
)
@staticmethod
def backward(ctx, dY):
X, W, m_sizes, gather_indices = ctx.saved_tensors
topk = ctx.topk
permute_x = ctx.permute_x
permute_y = ctx.permute_y
fuse_mul_post = ctx.fuse_mul_post
kernel_config_bwd_dX = ctx.kernel_config_bwd_dX
kernel_config_bwd_dW = ctx.kernel_config_bwd_dW
autotune = ctx.autotune
dX_only = ctx.dX_only
dW_only = ctx.dW_only
if not autotune:
if not dW_only:
assert kernel_config_bwd_dX is not None, (
"kernel_config_bwd_dX must be provided if autotune is False"
)
if not dX_only:
assert kernel_config_bwd_dW is not None, (
"kernel_config_bwd_dW must be provided if autotune is False"
)
assert not fuse_mul_post, (
"fused_mul should only be used for inference, not for training"
)
if not dX_only:
bwd_dW_config = {}
if kernel_config_bwd_dW is not None:
bwd_dW_config["use_tma_load_dy"] = kernel_config_bwd_dW.use_tma_load_dy
bwd_dW_config["use_tma_load_x"] = kernel_config_bwd_dW.use_tma_load_x
bwd_dW_config["use_tma_store"] = kernel_config_bwd_dW.use_tma_store
bwd_dW_config["BLOCK_SIZE_M"] = kernel_config_bwd_dW.BLOCK_SIZE_M
bwd_dW_config["BLOCK_SIZE_N"] = kernel_config_bwd_dW.BLOCK_SIZE_N
bwd_dW_config["BLOCK_SIZE_K"] = kernel_config_bwd_dW.BLOCK_SIZE_K
bwd_dW_config["num_warps"] = kernel_config_bwd_dW.num_warps
bwd_dW_config["num_stages"] = kernel_config_bwd_dW.num_stages
dW = grouped_gemm_dW(
X=X,
dY=dY,
m_sizes=m_sizes,
gather_indices=gather_indices,
topk=topk,
permute_x=permute_x,
permute_y=permute_y,
# Autotune -- this will override the manual kernel config if true
autotune=autotune,
# Manual kernel config
**bwd_dW_config,
)
else:
dW = None
if not dW_only:
bwd_dX_config = {}
if kernel_config_bwd_dX is not None:
bwd_dX_config["use_tma_load_dy"] = kernel_config_bwd_dX.use_tma_load_dy
bwd_dX_config["use_tma_load_w"] = kernel_config_bwd_dX.use_tma_load_w
bwd_dX_config["use_tma_store"] = kernel_config_bwd_dX.use_tma_store
bwd_dX_config["BLOCK_SIZE_M"] = kernel_config_bwd_dX.BLOCK_SIZE_M
bwd_dX_config["BLOCK_SIZE_N"] = kernel_config_bwd_dX.BLOCK_SIZE_N
bwd_dX_config["BLOCK_SIZE_K"] = kernel_config_bwd_dX.BLOCK_SIZE_K
bwd_dX_config["num_warps"] = kernel_config_bwd_dX.num_warps
bwd_dX_config["num_stages"] = kernel_config_bwd_dX.num_stages
dX = grouped_gemm_dX(
dY=dY,
W=W,
m_sizes=m_sizes,
gather_indices=gather_indices,
topk=topk,
permute_x=permute_x,
permute_y=permute_y,
# Autotune -- this will override the manual kernel config if true
autotune=autotune,
# Manual kernel config
**bwd_dX_config,
)
if topk > 1 and permute_x:
dX = dX.view(X.shape[0], topk, -1).sum(dim=1)
else:
dX = None
return (
dX,
dW,
None, # m_sizes
None, # gather_indices
None, # topk
None, # permute_x
None, # permute_y
None, # topk_weights
None, # fuse_mul_post
None, # kernel_config_fwd
None, # kernel_config_bwd_dX
None, # kernel_config_bwd_dW
None, # autotune
None, # dX_only
None, # dW_only
)
def check_valid_config_fwd(
permute_x,
permute_y,
use_tma_load_x,
use_tma_load_w,
use_tma_store,
fuse_mul_post,
is_first_gemm,
):
"""
Check if the configuration is valid for the forward pass.
"""
is_second_gemm = not is_first_gemm
assert not (permute_x and permute_y), "Cannot permute both X and Y"
assert not (is_second_gemm and permute_x), (
"Cannot permute X for the second grouped GEMM"
)
assert not (is_first_gemm and permute_y), (
"Cannot permute Y for the first grouped GEMM"
)
assert not (fuse_mul_post and is_first_gemm), (
"Cannot fuse mul for the first grouped GEMM"
)
assert not (use_tma_load_x and permute_x), (
"Cannot use TMA load and permute X unless on sm100+ (Blackwell+)"
)
assert not (use_tma_store and permute_y and is_second_gemm), (
"Cannot use TMA store and permute Y for the second grouped GEMM unless on sm100+ (Blackwell+)"
)
def check_valid_config_bwd_dW(
permute_x,
permute_y,
use_tma_load_dY,
use_tma_load_x,
use_tma_store,
fuse_mul_post,
is_first_gemm,
):
"""
Check if the configuration is valid for the backward pass of dW.
"""
is_second_gemm = not is_first_gemm
if fuse_mul_post:
assert False, "Cannot fuse_mul is not supported for backward pass"
if is_second_gemm and permute_y and use_tma_load_dY:
assert False, "Cannot use TMA load and permute Y for the second grouped GEMM"
if is_first_gemm and permute_x and use_tma_load_x:
assert False, "Cannot use TMA load and permute X for the first grouped GEMM"
def check_valid_config_bwd_dX(
permute_x,
permute_y,
use_tma_load_dY,
use_tma_load_w,
use_tma_store,
fuse_mul_post,
is_first_gemm,
):
"""
Check if the configuration is valid for the backward pass of dW.
"""
is_second_gemm = not is_first_gemm
if fuse_mul_post:
assert False, "Cannot fuse_mul is not supported for backward pass"
if is_second_gemm and permute_y and use_tma_load_dY:
assert False, "Cannot use TMA load and permute Y for the second grouped GEMM"
if use_tma_store and permute_x and is_first_gemm:
assert False, "Cannot use TMA store and permute X for the first grouped GEMM"
def grouped_gemm(
X: torch.Tensor,
W: torch.Tensor,
m_sizes: torch.Tensor,
topk: int,
gather_indices: torch.Tensor = None,
permute_x: bool = False,
permute_y: bool = False,
topk_weights=None,
fuse_mul_post=False,
kernel_config_fwd: KernelConfigForward = None,
kernel_config_bwd_dX: KernelConfigBackward_dX = None,
kernel_config_bwd_dW: KernelConfigBackward_dW = None,
autotune: bool = False,
is_first_gemm: bool = True,
# Only for debugging
dX_only: bool = False,
dW_only: bool = False,
):
"""
Grouped GEMM for MoE MLPs.
The implementation offers a number of fusions specific to MoE:
- `permute_x`: fuse the permutation of hidden states from token order (original order) to grouped expert order, typically only needed for the first grouped GEMM in an MoE MLP.
- When `permute_x` is True, `X` is expected to be of shape (num_tokens, K).
- When `permute_x` is False, `X` is expected to be of shape (total_tokens, K) where `total_tokens = num_tokens * topk` AND already permuted to grouped expert order, i.e., hidden states are sorted such that tokens assigned to each expert are contiguous.
- `permute_y`: fused the permuation of the output from expert grouped order back to original token order, typically only needed for the second grouped GEMM in an MoE MLP.
- `fuse_mul`: fuse the multiplication of the routed output with topk_weights, used only when `permute_y` is True. NOTE: this should only be used when using this kernel for inference, not for training.
X: (M, K) hidden states where M is the num_tokens if `permute_x` is True, otherwise `total_tokens` where `total_tokens = num_tokens * topk`.
W: (E, N, K) expert weights, where E is number of experts, N in the intermediate (output) dim, and K is the reduction dim
m_sizes: tokens assigned to each expert which correspond to the size of M in the respective GEMMs in the grouped GEMM.
gather_indices: (total_tokens,) indices of tokens assigned to each expert. E.g., slicing gather_indices by cumsum of m_sizes gives the indices of tokens assigned to each expert. Needed when either `permute_x` or `permute_y` is True.
topk_weights: (total_tokens,) weights to multiply routed output by in expert MLP calculation, used only when `fuse_mul` is True (see note on `fuse_mul`).
kernel_config_fwd: KernelConfigForward for forward pass.
kernel_config_bwd_dX: KernelConfigBackward_dX for backward pass of dX.
kernel_config_bwd_dW: KernelConfigBackward_dW for backward pass of dW.
autotune: whether to autotune the kernel, if yes, kernel_config_fwd, kernel_config_bwd_dX, and kernel_config_bwd_dW will be ignored.
is_first_gemm: whether this is the first grouped GEMM in an MoE MLP. This is needed to check whether kernel configs are valid. `permute_x` should only be used for first gemm; `permute_y` should only be used for second gemm.
This will impact whether TMA can be used for loading and storing.
"""
if not autotune:
assert kernel_config_fwd is not None, (
"kernel_config_fwd must be provided if autotune is False"
)
check_valid_config_fwd(
permute_x,
permute_y,
use_tma_load_x=kernel_config_fwd.use_tma_load_x,
use_tma_load_w=kernel_config_fwd.use_tma_load_w,
use_tma_store=kernel_config_fwd.use_tma_store,
fuse_mul_post=fuse_mul_post,
is_first_gemm=is_first_gemm,
)
if kernel_config_bwd_dW is not None and not dX_only:
check_valid_config_bwd_dW(
permute_x,
permute_y,
use_tma_load_dY=kernel_config_bwd_dW.use_tma_load_dy,
use_tma_load_x=kernel_config_bwd_dW.use_tma_load_x,
use_tma_store=kernel_config_bwd_dW.use_tma_store,
fuse_mul_post=fuse_mul_post,
is_first_gemm=is_first_gemm,
)
if kernel_config_bwd_dX is not None and not dW_only:
check_valid_config_bwd_dX(
permute_x,
permute_y,
use_tma_load_dY=kernel_config_bwd_dX.use_tma_load_dy,
use_tma_load_w=kernel_config_bwd_dX.use_tma_load_w,
use_tma_store=kernel_config_bwd_dX.use_tma_store,
fuse_mul_post=fuse_mul_post,
is_first_gemm=is_first_gemm,
)
if permute_x or permute_y:
assert gather_indices is not None, (
"gather_indices is required when either permute_x or permute_y is True"
)
if fuse_mul_post:
assert topk_weights is not None, (
"topk_weights is required when fuse_mul_post is True"
)
X = X.view(-1, X.shape[-1])
m_sizes = m_sizes.view(-1)
gather_indices = gather_indices.view(-1)
return GroupedGemm.apply(
X,
W,
m_sizes,
topk,
gather_indices,
permute_x,
permute_y,
topk_weights,
fuse_mul_post,
kernel_config_fwd,
kernel_config_bwd_dX,
kernel_config_bwd_dW,
autotune,
dX_only,
dW_only,
)

View file

@ -0,0 +1,393 @@
"""
Autotuning utils
"""
import logging
from itertools import product
from typing import List
import torch
import triton
logger = logging.getLogger(__name__)
DEFAULT_M_BLOCK_SIZES = [64, 128]
DEFAULT_N_BLOCK_SIZES = [64, 128, 256]
DEFAULT_K_BLOCK_SIZES = [64, 128, 256]
DEFAULT_NUM_CTAS = 1
DEFAULT_NUM_WARPS = [4, 8]
DEFAULT_NUM_STAGES = [3, 4, 5]
BOOLS = [True, False]
def val_to_list(val):
if val is None:
return None
elif isinstance(val, list):
return val
else:
return [val]
def convert_args_to_list(args):
return [val_to_list(arg) for arg in args]
def get_forward_configs(
BLOCK_M=DEFAULT_M_BLOCK_SIZES,
BLOCK_N=DEFAULT_N_BLOCK_SIZES,
BLOCK_K=DEFAULT_K_BLOCK_SIZES,
TMA_LOAD_X=True,
TMA_LOAD_W=True,
TMA_STORE=False, # NOTE: TMA_STORE is disabled for now
num_warps=DEFAULT_NUM_WARPS,
num_stages=DEFAULT_NUM_STAGES,
num_ctas=DEFAULT_NUM_CTAS,
):
(
BLOCK_M,
BLOCK_N,
BLOCK_K,
TMA_LOAD_X,
TMA_LOAD_W,
TMA_STORE,
num_warps,
num_stages,
num_ctas,
) = convert_args_to_list(
[
BLOCK_M,
BLOCK_N,
BLOCK_K,
TMA_LOAD_X,
TMA_LOAD_W,
TMA_STORE,
num_warps,
num_stages,
num_ctas,
]
)
kernel_configs = []
for (
block_m,
block_n,
block_k,
w,
s,
tma_load_x,
tma_load_w,
tma_store,
num_ctas,
) in product(
BLOCK_M,
BLOCK_N,
BLOCK_K,
num_warps,
num_stages,
TMA_LOAD_X,
TMA_LOAD_W,
TMA_STORE,
num_ctas,
):
kernel_configs.append(
triton.Config(
dict(
BLOCK_SIZE_M=block_m,
BLOCK_SIZE_N=block_n,
BLOCK_SIZE_K=block_k,
USE_TMA_LOAD_X=tma_load_x,
USE_TMA_LOAD_W=tma_load_w,
USE_TMA_STORE=tma_store,
),
num_warps=w,
num_stages=s,
num_ctas=num_ctas,
)
)
return kernel_configs
def get_dX_kernel_configs(
BLOCK_M=DEFAULT_M_BLOCK_SIZES,
BLOCK_N=DEFAULT_N_BLOCK_SIZES,
BLOCK_K=DEFAULT_K_BLOCK_SIZES,
TMA_LOAD_dY=True,
TMA_LOAD_W=True,
TMA_STORE=False, # NOTE: TMA_STORE is disabled for now
num_warps=DEFAULT_NUM_WARPS,
num_stages=DEFAULT_NUM_STAGES,
num_ctas=DEFAULT_NUM_CTAS,
):
(
BLOCK_M,
BLOCK_N,
BLOCK_K,
TMA_LOAD_dY,
TMA_LOAD_W,
TMA_STORE,
num_warps,
num_stages,
num_ctas,
) = convert_args_to_list(
[
BLOCK_M,
BLOCK_N,
BLOCK_K,
TMA_LOAD_dY,
TMA_LOAD_W,
TMA_STORE,
num_warps,
num_stages,
num_ctas,
]
)
kernel_configs = []
for (
block_m,
block_n,
block_k,
w,
s,
tma_load_dy,
tma_load_w,
tma_store,
num_ctas,
) in product(
BLOCK_M,
BLOCK_N,
BLOCK_K,
num_warps,
num_stages,
TMA_LOAD_dY,
TMA_LOAD_W,
TMA_STORE,
num_ctas,
):
kernel_configs.append(
triton.Config(
dict(
BLOCK_SIZE_M=block_m,
BLOCK_SIZE_N=block_n,
BLOCK_SIZE_K=block_k,
USE_TMA_LOAD_dY=tma_load_dy,
USE_TMA_LOAD_W=tma_load_w,
USE_TMA_STORE=tma_store,
),
num_warps=w,
num_stages=s,
num_ctas=num_ctas,
)
)
return kernel_configs
def get_dW_kernel_configs(
BLOCK_M=DEFAULT_M_BLOCK_SIZES,
BLOCK_N=DEFAULT_N_BLOCK_SIZES,
BLOCK_K=DEFAULT_K_BLOCK_SIZES,
num_warps=DEFAULT_NUM_WARPS,
num_stages=DEFAULT_NUM_STAGES,
num_ctas=DEFAULT_NUM_CTAS,
TMA_LOAD_dY=True,
TMA_LOAD_X=True,
TMA_STORE=False,
):
(
BLOCK_M,
BLOCK_N,
BLOCK_K,
num_warps,
num_stages,
num_ctas,
TMA_LOAD_dY,
TMA_LOAD_X,
TMA_STORE,
) = convert_args_to_list(
[
BLOCK_M,
BLOCK_N,
BLOCK_K,
num_warps,
num_stages,
num_ctas,
TMA_LOAD_dY,
TMA_LOAD_X,
TMA_STORE,
]
)
kernel_configs = []
for (
block_m,
block_n,
block_k,
w,
s,
tma_load_dy,
tma_load_x,
tma_store,
num_ctas,
) in product(
BLOCK_M,
BLOCK_N,
BLOCK_K,
num_warps,
num_stages,
TMA_LOAD_dY,
TMA_LOAD_X,
TMA_STORE,
num_ctas,
):
kernel_configs.append(
triton.Config(
dict(
BLOCK_SIZE_M=block_m,
BLOCK_SIZE_N=block_n,
BLOCK_SIZE_K=block_k,
USE_TMA_LOAD_dY=tma_load_dy,
USE_TMA_LOAD_X=tma_load_x,
USE_TMA_STORE=tma_store,
),
num_warps=w,
num_stages=s,
num_ctas=num_ctas,
)
)
return kernel_configs
def estimate_smem_reqs(
num_stages: int,
BLOCK_SIZE_M: int,
BLOCK_SIZE_N: int,
BLOCK_SIZE_K: int,
dtype: torch.dtype,
):
num_bytes = dtype.itemsize
return (
num_stages * BLOCK_SIZE_K * (BLOCK_SIZE_M + BLOCK_SIZE_N)
+ BLOCK_SIZE_M * BLOCK_SIZE_N
) * num_bytes
def exceeds_smem_capacity(
num_stages: int,
BLOCK_SIZE_M: int,
BLOCK_SIZE_N: int,
BLOCK_SIZE_K: int,
dtype: torch.dtype,
smem_size: int,
slack: float = 50000,
):
smem_reqs = estimate_smem_reqs(
num_stages, BLOCK_SIZE_M, BLOCK_SIZE_N, BLOCK_SIZE_K, dtype
)
return smem_reqs > smem_size + slack
def common_prune_criteria(config: triton.Config, kwargs: dict, dtype):
from grouped_gemm.interface import supports_tma
from grouped_gemm.kernels.tuning import get_device_properties
smem_size = get_device_properties().SIZE_SMEM
num_stages = config.num_stages
BLOCK_SIZE_M = config.kwargs["BLOCK_SIZE_M"]
BLOCK_SIZE_N = config.kwargs["BLOCK_SIZE_N"]
BLOCK_SIZE_K = config.kwargs["BLOCK_SIZE_K"]
num_tokens = kwargs["NUM_TOKENS"]
num_experts = kwargs["NUM_EXPERTS"]
permute_x = kwargs["PERMUTE_X"]
permute_y = kwargs["PERMUTE_Y"]
tokens_per_expert = num_tokens // num_experts
# use_tma = [k for k in config.kwargs.keys() if k.startswith("USE_TMA_")]
MIN_BLOCK_SIZE_M = DEFAULT_M_BLOCK_SIZES[0]
if exceeds_smem_capacity(
num_stages, BLOCK_SIZE_M, BLOCK_SIZE_N, BLOCK_SIZE_K, dtype, smem_size
):
return True
if BLOCK_SIZE_M > tokens_per_expert * 2 and tokens_per_expert > MIN_BLOCK_SIZE_M:
return True
if permute_x and permute_y:
return True
# if not supports_tma() and any(use_tma):
# return True
return False
def maybe_disable_tma(config: triton.Config):
from grouped_gemm.interface import supports_tma
tma_keys = [k for k in config.kwargs.keys() if k.startswith("USE_TMA_")]
if not supports_tma():
logger.info("Disabling TMA")
for k in tma_keys:
config.kwargs[k] = False
def prune_kernel_configs_fwd(configs: list[triton.Config], args, **kwargs):
x = kwargs["x_ptr"]
dtype = x.dtype
logger.debug(f"Pruning configs: {len(configs)}")
pruned_configs = []
for config in configs:
# disable TMA if gpu does not support it
maybe_disable_tma(config)
if common_prune_criteria(config, kwargs, dtype):
continue
if config.kwargs["USE_TMA_LOAD_X"] and kwargs["PERMUTE_X"]:
# Dynamically disable TMA_LOAD_X for permuted X
config.kwargs["USE_TMA_LOAD_X"] = False
if config.kwargs["USE_TMA_STORE"] and kwargs["PERMUTE_Y"]:
continue
pruned_configs.append(config)
logger.debug(f"Pruned configs: {len(pruned_configs)}")
return pruned_configs
def prune_dX_configs(configs: List[triton.Config], args, **kwargs):
dtype = kwargs["w_ptr"].dtype
logger.debug(f"Pruning configs: {len(configs)}")
pruned_configs = []
for config in configs:
if common_prune_criteria(config, kwargs, dtype):
continue
if config.kwargs["USE_TMA_LOAD_dY"] and kwargs["PERMUTE_Y"]:
# dynamically disable TMA_LOAD_dY for permuted Y
config.kwargs["USE_TMA_LOAD_dY"] = False
if config.kwargs["USE_TMA_STORE"] and kwargs["PERMUTE_X"]:
continue
pruned_configs.append(config)
logger.debug(f"Pruned configs: {len(pruned_configs)}")
return pruned_configs
def prune_kernel_configs_backward_dW(configs: list[triton.Config], args, **kwargs):
dtype = kwargs["x_ptr"].dtype
pruned_configs = []
logger.debug(f"Pruning configs: {len(configs)}")
for config in configs:
if common_prune_criteria(config, kwargs, dtype):
continue
if config.kwargs["USE_TMA_LOAD_dY"] and kwargs["PERMUTE_Y"]:
config.kwargs["USE_TMA_LOAD_dY"] = False
if config.kwargs["USE_TMA_LOAD_X"] and kwargs["PERMUTE_X"]:
config.kwargs["USE_TMA_LOAD_X"] = False
pruned_configs.append(config)
logger.debug(f"Pruned configs: {len(pruned_configs)}")
return pruned_configs

View file

@ -0,0 +1,499 @@
import torch
import triton
import triton.language as tl
from grouped_gemm.kernels.autotuning import (
get_dW_kernel_configs,
get_dX_kernel_configs,
prune_dX_configs,
prune_kernel_configs_backward_dW,
)
"""
dX backward kernel
- Shapes
- the forward pass input X shape is [NUM_TOKENS, K] if permute_x else [NUM_TOKENS * TOPK, K]; output y is [NUM_TOKENS * TOPK, N]
- the backward pass input dy shape is [NUM_TOKENS * TOPK, N], reduce across N, output dX is [NUM_TOKENS * TOPK, K]
- Note that in the backward pass, the output size is still [NUM_TOKENS * TOPK, K] since we still need to accumulate gradients for each expert chosen by the token in a post-processing step.
`permute_x` notes:
- In the forward pass, if we permute X on load, we need to permute on store in the backward pass to restore to original token order
- the output dX with have shape [NUM_TOKENS * TOPK, K] and we need to perform an additional reduction across topk to accumulate gradients
- This is done as a post-processing step in autograd.Function.
- If not `permute_x`, this postprocessing step should take place outside autograd.Function such that the gradient shape matches the input X shape.
`permute_y` notes:
- In the forward pass, if we permuted output on store (e.g., in the second grouped GEMM in fused MoE MLP), we need to permute on load to get from token order to expert grouped order
- We still store in contiguous order since we are writing out dX which will be the input to the backwards pass of the first grouped GEMM
`fused_mul` notes:
- In the forward pass, if we used the multiplication of topk weights (e.g., in the second grouped GEMM in fused MoE MLP), we need to make a few additional changes:
1) We load topk_weights in natural (token) order. Since we only enable `fuse_mul` when permuting on store (`permute_y`), we multiply grad_output by topk_weights before backpropagating
2) We need to calculate the gradient of the topk_weights. This gets messy since we need do an additioanl elementwise multiplication in the GEMM main loop and then write out in unpermuted order. For now, we do not fuse this step but calculate as a simple
Invalid combinations:
- permute_y and use_tma_load: permuting y on store in forward -> load in permuted order in backward, therefore can't use TMA load (unless Blackwell which supports gather / scatter TMA)
- permute_x and use_tma_store: permuting x on load in forward -> store in permuted order in backward, therefore can't use TMA store (unless Blackwell which supports gather / scatter TMA)
TODO:
- We define indices for all conditions and expect that unused indices will be DCE'd during compilation. Check that this is the case otherwise will result in unnecessary register usage.
"""
@triton.jit
def _grouped_gemm_dX_kernel(
dY_ptr, # [M_total, N]
w_ptr, # [E, N, K]
dX_ptr, # [M_total, K]
gather_indices_ptr,
m_sizes_ptr,
# problem sizes
NUM_EXPERTS: tl.constexpr,
NUM_TOKENS: tl.constexpr,
TOPK: tl.constexpr,
N: tl.constexpr,
K: tl.constexpr,
NUM_SMS: tl.constexpr,
# Tuning parameters
BLOCK_SIZE_M: tl.constexpr,
BLOCK_SIZE_N: tl.constexpr,
BLOCK_SIZE_K: tl.constexpr,
PERMUTE_X: tl.constexpr = False,
PERMUTE_Y: tl.constexpr = False,
USE_TMA_LOAD_W: tl.constexpr = False,
USE_TMA_LOAD_dY: tl.constexpr = False,
USE_TMA_STORE: tl.constexpr = False,
FLATTEN: tl.constexpr = True,
) -> None:
TOTAL_TOKENS: tl.constexpr = NUM_TOKENS * TOPK
output_dtype = dX_ptr.dtype.element_ty
tidx = tl.program_id(0)
# This removes the need for predication along N in the GEMM main loop
tl.static_assert(N % BLOCK_SIZE_N == 0, "N must be divisible by BLOCK_SIZE_N")
tl.static_assert(K % BLOCK_SIZE_K == 0, "K must be divisible by BLOCK_SIZE_K")
# Create TMA descriptors for loading sorted tokens
# When using TMA load, we don't permute_x, so shape should be [TOTAL_TOKENS, K]
# Also, we are defining a single global descriptor with single block shape
# Need to check that this does not result in errors when crossing expert boundaries
if USE_TMA_LOAD_dY:
dY_desc = tl._experimental_make_tensor_descriptor(
dY_ptr,
shape=[TOTAL_TOKENS, N],
strides=[N, 1],
block_shape=[BLOCK_SIZE_M, BLOCK_SIZE_N],
)
if USE_TMA_LOAD_W:
expert_stride = N * K
w_desc = tl._experimental_make_tensor_descriptor(
w_ptr,
shape=[NUM_EXPERTS, N, K],
strides=[expert_stride, K, 1],
block_shape=[1, BLOCK_SIZE_N, BLOCK_SIZE_K],
)
m_end = 0
processed_tiles = 0
m_block_range = tl.arange(0, BLOCK_SIZE_M)
n_block_range = tl.arange(0, BLOCK_SIZE_N)
k_block_range = tl.arange(0, BLOCK_SIZE_K)
for expert_idx in range(NUM_EXPERTS, flatten=FLATTEN):
m_start = m_end
m_size = tl.load(m_sizes_ptr + expert_idx).to(tl.int32)
m_end = m_start + m_size
if m_size > 0:
# Advance n offset to the weights for that respective expert
n_start = expert_idx * N
# N_start_offset = g.to(tl.int64) * N
# tiles for this group's GEMM
num_m_tiles = tl.cdiv(m_size, BLOCK_SIZE_M)
num_k_tiles = tl.cdiv(K, BLOCK_SIZE_K)
num_tiles_per_expert = num_m_tiles * num_k_tiles
if USE_TMA_STORE:
# Need to define descript within loop to predicate store along M
tl.static_assert(
K % BLOCK_SIZE_K == 0, "K must be divisible by BLOCK_SIZE_K"
)
dX_desc = tl._experimental_make_tensor_descriptor(
dX_ptr,
shape=[m_end, K],
strides=[K, 1],
block_shape=[BLOCK_SIZE_M, BLOCK_SIZE_K],
)
# Lower bound and upper bound are defined relative to the total tiles processed so far
# This ensures that we are only processing tiles for the current expert group AND
# we never exceed the total number of tiles for all expert groups
while tidx >= processed_tiles and tidx < (
processed_tiles + num_tiles_per_expert
):
group_index = tidx - processed_tiles
# Output tile for this thread block for this expert group
tile_m_idx = group_index % num_m_tiles
tile_k_idx = group_index // num_m_tiles
if PERMUTE_X or PERMUTE_Y:
# These will be used for loading and storing in permuted order
gather_offsets = tile_m_idx * BLOCK_SIZE_M + m_block_range
# indices_to_gather = m_start + gather_offsets
indices_to_gather = m_start + tl.max_contiguous(
tl.multiple_of(gather_offsets % m_size, BLOCK_SIZE_M),
BLOCK_SIZE_M,
)
expert_token_idx = tl.load(
gather_indices_ptr + indices_to_gather,
mask=indices_to_gather < TOTAL_TOKENS,
)
expert_token_offsets = expert_token_idx[:, None]
# Masks for permuted load and store
row_mask = gather_offsets < m_size
row_mask = row_mask[:, None]
# We only take into account the following two cases: (PERMUTE_X and NOT PERMUTE_Y) and (NOT PERMUTE_X and PERMUTE_Y)
# Hence, we can make the following simplifying assumptions when loading and storing
# Note the different strides between the two cases: the offsets for loading and storing are flipped and the strides must also be adjusted
if PERMUTE_X:
# Case where we permuted on load in the forward pass (typically first grouped GEMM in MoE MLP)
load_a_idx = (
indices_to_gather[:, None] * N
) # Load in contiguous (expert grouped) order
store_idx = (
expert_token_offsets * K
) # Permute on store from expert -> token order
else:
# Case where we permuted on store in the forward pass (typically second grouped GEMM in MoE MLP)
load_a_idx = (
expert_token_offsets * N
) # Permute on load from token -> expert order
store_idx = (
indices_to_gather[:, None] * K
) # Store in contiguous order
else:
# # Position in full matrix - needed for TMA
# m_offset = (M_start + (tile_m_idx * BLOCK_SIZE_M)).to(tl.int32)
# k_offset = (tile_k_idx * BLOCK_SIZE_K).to(tl.int32)
# Offsets *relative* to the *current* expert -- m_start will then advance to this expert's start token
offs_am = tile_m_idx * BLOCK_SIZE_M + m_block_range
# [M, N] @ [N, K] -> [M, K] => Stride for A is N, stride for B is K
# We need two additional offsets:
# 1. For A, m_start to advance to this expert's start token
# 2. For B, n_start to advance to this expert's weights since we are passing in an [E, N, K] weight matrix
row_offsets_a = m_start + offs_am[:, None]
load_a_idx = row_offsets_a * N
store_idx = row_offsets_a * K
row_mask = offs_am[:, None] < m_size
if not USE_TMA_LOAD_dY:
dY_ptrs = dY_ptr + load_a_idx + n_block_range[None, :]
offs_bk = tile_k_idx * BLOCK_SIZE_K + k_block_range
if not USE_TMA_LOAD_W:
row_offsets_b = n_start + n_block_range
# offs_bn = n_start + n_block_range
# row_offsets_b = tl.max_contiguous(tl.multiple_of(offs_bn, BLOCK_SIZE_N), BLOCK_SIZE_N)
w_ptrs = w_ptr + row_offsets_b[:, None] * K + offs_bk[None, :]
# TODO: check whether predication along K is needed since we checked that K is divisible by BLOCK_SIZE_K in the forward kernel
# col_mask = offs_bk[None, :] < K
store_mask = row_mask # & col_mask
accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_K), dtype=tl.float32)
# GEMM main loop
for n_offset in range(0, N, BLOCK_SIZE_N):
# dY block [M, N]
if not USE_TMA_LOAD_dY:
dY = tl.load(dY_ptrs, mask=row_mask)
else:
dY = dY_desc.load(
[m_start + tile_m_idx * BLOCK_SIZE_M, n_offset]
)
if not USE_TMA_LOAD_W:
w = tl.load(w_ptrs) # , mask=col_mask)
else:
w = w_desc.load(
[expert_idx, n_offset, tile_k_idx * BLOCK_SIZE_K]
)
w = tl.reshape(w, (BLOCK_SIZE_N, BLOCK_SIZE_K))
# TODO: check if predication along K is needed since we checked that K is divisible by BLOCK_SIZE_K in the forward kernel
# [M, N] @ [N, K] -> [M, K]
accumulator += tl.dot(dY, w) # NOTE: no transpose of b
# Advance A along contiguous dimension
if not USE_TMA_LOAD_dY:
dY_ptrs += BLOCK_SIZE_N
# Note we are no longer advancing B along contiguous dimension since weights are arranged as [N, K]
# Instead, we need to stride by K to advance to the [N_BLOCK_SIZE, K_BLOCK_SIZE] tile
if not USE_TMA_LOAD_W:
w_ptrs += BLOCK_SIZE_N * K
dX = accumulator.to(output_dtype)
# Writing out a BLOCK_M x BLOCK_K tile, so we need to stride by K
if USE_TMA_STORE:
offset_m = tile_m_idx * BLOCK_SIZE_M # .to(tl.int32)
offset_k = tile_k_idx * BLOCK_SIZE_K # .to(tl.int32)
dX_desc.store([m_start + offset_m, offset_k], dX)
else:
tl.store(
dX_ptr + store_idx + offs_bk[None, :],
dX,
mask=store_mask,
)
# Move to the next tile within this expert group
tidx += NUM_SMS
# Update the total tiles count for the next expert group
processed_tiles += num_tiles_per_expert
_autotuned_grouped_gemm_dX_kernel = triton.autotune(
configs=get_dX_kernel_configs(),
prune_configs_by={"early_config_prune": prune_dX_configs},
key=["NUM_EXPERTS", "NUM_TOKENS", "N", "K", "PERMUTE_X", "PERMUTE_Y"],
)(_grouped_gemm_dX_kernel)
"""
notes on permute_x:
- for the first grouped GEMM, we permuted on load -> X was [num_tokens, K] and stored y in expert grouped order [num_tokens * topk, K]
- in the backwards pass, we need to permute on load of X while loading dy in contiguous (expert grouped) order
- since we are writing out dW, there is no need to permute on store
notes on permute_y:
- for the second grouped GEMM, we permuted on store -> y was permuted from expert grouped order to token order, x was loaded in expert grouped order since it was the output of the first grouped GEMM
- in the backwards pass, we need to permute on load of dy to get from token order to expert grouped order to match the order of X
- since we are writing out dW, there is no need to permute on store
notes on TMA loading:
- if we're TMA loading both X and dY, then we need to mask along the M dimension
to account for expert boundaries
- we can either
- define TMA descriptors within the outer for loop to predicate loads
or
- mask along M after loading
"""
@triton.jit
def _grouped_gemm_dW_kernel(
x_ptr,
dY_ptr,
dW_ptr,
m_sizes_ptr,
gather_indices_ptr,
# problem sizes
NUM_TOKENS: tl.constexpr,
TOPK: tl.constexpr,
NUM_EXPERTS: tl.constexpr,
N: tl.constexpr,
K: tl.constexpr,
NUM_SMS: tl.constexpr,
BLOCK_SIZE_N: tl.constexpr,
BLOCK_SIZE_K: tl.constexpr,
BLOCK_SIZE_M: tl.constexpr,
PERMUTE_X: tl.constexpr = False,
PERMUTE_Y: tl.constexpr = False,
USE_TMA_LOAD_dY: tl.constexpr = False,
USE_TMA_LOAD_X: tl.constexpr = False,
USE_TMA_STORE: tl.constexpr = False,
FLATTEN: tl.constexpr = True,
acc_dtype: tl.constexpr = tl.float32,
) -> None:
TOTAL_TOKENS: tl.constexpr = NUM_TOKENS * TOPK
TMA_LOAD_BOTH: tl.constexpr = USE_TMA_LOAD_X and USE_TMA_LOAD_dY
tidx = tl.program_id(0)
output_dtype = dW_ptr.dtype.element_ty
if USE_TMA_LOAD_dY and not TMA_LOAD_BOTH:
dY_desc = tl._experimental_make_tensor_descriptor(
dY_ptr,
shape=[TOTAL_TOKENS, N],
strides=[N, 1],
block_shape=[BLOCK_SIZE_M, BLOCK_SIZE_N],
)
if USE_TMA_LOAD_X and not TMA_LOAD_BOTH:
x_desc = tl._experimental_make_tensor_descriptor(
x_ptr,
shape=[TOTAL_TOKENS, K],
strides=[K, 1],
block_shape=[BLOCK_SIZE_M, BLOCK_SIZE_K],
)
# Output tiles per expert, since each expert weight matrix is [N, K]
num_n_tiles = tl.cdiv(N, BLOCK_SIZE_N)
num_k_tiles = tl.cdiv(K, BLOCK_SIZE_K)
output_tiles_per_expert = num_n_tiles * num_k_tiles
block_range_m = tl.arange(0, BLOCK_SIZE_M)
block_range_n = tl.arange(0, BLOCK_SIZE_N)
block_range_k = tl.arange(0, BLOCK_SIZE_K)
# NOTE: Important that N % BLOCK_SIZE_N == 0 and K % BLOCK_SIZE_K == 0 when using TMA store
if USE_TMA_STORE:
tl.static_assert(N % BLOCK_SIZE_N == 0, "N must be divisible by BLOCK_SIZE_N")
tl.static_assert(K % BLOCK_SIZE_K == 0, "K must be divisible by BLOCK_SIZE_K")
dW_desc = tl._experimental_make_tensor_descriptor(
dW_ptr,
shape=[NUM_EXPERTS, N, K],
strides=[N * K, K, 1],
block_shape=[1, BLOCK_SIZE_N, BLOCK_SIZE_K],
)
for tile_idx in range(
tidx, output_tiles_per_expert, NUM_SMS
): # , flatten=FLATTEN):
# Output tile index
tile_n_idx = tile_idx % num_n_tiles
tile_k_idx = tile_idx // num_n_tiles
# Output tile offsets
n_offset = tile_n_idx * BLOCK_SIZE_N
k_offset = tile_k_idx * BLOCK_SIZE_K
# For storing
# TODO: Check whether the k mask is needed since we statically check that K is divisible by BLOCK_SIZE_K in the forward kernel
# ditto for n_mask
n_mask = block_range_n + n_offset < N
k_mask = block_range_k + k_offset < K
nk_mask = n_mask[:, None] & k_mask[None, :]
m_end = 0
for expert_idx in range(NUM_EXPERTS):
# We need to instantiate a fresh accumulator for each expert
accumulator = tl.zeros((BLOCK_SIZE_N, BLOCK_SIZE_K), dtype=acc_dtype)
m_start = m_end
# Need to figure out why this cast is needed, otherwise compiler complains about mismatching types
m_size = tl.load(m_sizes_ptr + expert_idx).to(tl.int32)
m_end = m_start + m_size
# NOTE: when storing the result, we need to offset by n_start since we are storing the result for this expert to the global [E, N, K] weight matrix
n_start = expert_idx * N
store_row_offs = n_start + n_offset + block_range_n
if m_size > 0:
if TMA_LOAD_BOTH:
dY_desc = tl._experimental_make_tensor_descriptor(
dY_ptr,
shape=[m_end, N],
strides=[N, 1],
block_shape=[BLOCK_SIZE_M, BLOCK_SIZE_N],
)
x_desc = tl._experimental_make_tensor_descriptor(
x_ptr,
shape=[m_end, K],
strides=[K, 1],
block_shape=[BLOCK_SIZE_M, BLOCK_SIZE_K],
)
for tile_m_idx in range(0, m_size, BLOCK_SIZE_M):
m_block_size = tl.minimum(BLOCK_SIZE_M, m_size - tile_m_idx)
if m_block_size > 0:
# Global offset for this chunk
m_global_offset = m_start + tile_m_idx
m_offsets = m_global_offset + block_range_m
if PERMUTE_X or PERMUTE_Y:
# These will be used for loading and storing in permuted order
gather_offsets = (
tile_m_idx + block_range_m
) # NOTE: tile_m_idx is already strided by BLOCK_SIZE_M
indices_to_gather = m_start + tl.max_contiguous(
tl.multiple_of(gather_offsets % m_size, BLOCK_SIZE_M),
BLOCK_SIZE_M,
)
# indices_to_gather = m_start + gather_offsets
expert_token_idx = tl.load(
gather_indices_ptr + indices_to_gather,
mask=indices_to_gather < TOTAL_TOKENS,
)
expert_token_offsets = expert_token_idx[:, None]
# Masks for permuted load and store
row_load_mask = gather_offsets < m_size
# We only take into account the following two cases: (PERMUTE_X and NOT PERMUTE_Y) and (NOT PERMUTE_X and PERMUTE_Y)
# Hence, we can make the following simplifying assumptions when loading and storing
# Note the different strides between the two cases: the offsets for loading and storing are flipped and the strides must also be adjusted
if PERMUTE_X:
x_row_load_idx = (
(expert_token_offsets // TOPK) * K
) # Permute on load from token -> expert order, divide by TOPK to index from original number of tokens
dY_row_load_idx = m_offsets[:, None] * N
else:
x_row_load_idx = (
indices_to_gather[:, None] * K
) # Load in contiguous order (no permutation on load)
dY_row_load_idx = expert_token_offsets * N
else:
x_row_load_idx = m_offsets[:, None] * K
dY_row_load_idx = m_offsets[:, None] * N
row_load_mask = block_range_m < m_block_size
mk_mask = row_load_mask[:, None] & k_mask[None, :]
mn_mask = row_load_mask[:, None] & n_mask[None, :]
if USE_TMA_LOAD_X:
x = x_desc.load([m_global_offset, k_offset])
else:
x = tl.load(
x_ptr
+ x_row_load_idx
+ (k_offset + block_range_k)[None, :],
mask=mk_mask,
)
if USE_TMA_LOAD_dY:
dY = dY_desc.load([m_global_offset, n_offset])
else:
dY = tl.load(
dY_ptr
+ dY_row_load_idx
+ (n_offset + block_range_n)[None, :],
mask=mn_mask,
)
accumulator += tl.dot(
dY.T, # [BLOCK_N, BLOCK_M]
x, # [BLOCK_M, BLOCK_K]
)
y = accumulator.to(output_dtype)
if USE_TMA_STORE:
# Need to expand dims to match [E, N, K] shape
y = tl.expand_dims(y, 0)
dW_desc.store([expert_idx, n_offset, k_offset], y)
else:
tl.store(
dW_ptr
# + (n_offset + offs_n)[:, None] * K
+ store_row_offs[:, None] * K
+ (k_offset + block_range_k)[None, :],
y,
mask=nk_mask,
)
_autotuned_grouped_gemm_dW_kernel = triton.autotune(
configs=get_dW_kernel_configs(),
prune_configs_by={"early_config_prune": prune_kernel_configs_backward_dW},
key=["NUM_EXPERTS", "NUM_TOKENS", "N", "K", "PERMUTE_X", "PERMUTE_Y"],
)(_grouped_gemm_dW_kernel)

View file

@ -0,0 +1,262 @@
import torch
import triton
import triton.language as tl
from grouped_gemm.kernels.autotuning import (
get_forward_configs,
prune_kernel_configs_fwd,
)
#
# PERMUTE_X -> permute tokens so that they are ordered by expert
# PERMUTE_Y -> permute output so that they are ordered by token
# These are effectively the same thing: the former loads in permuted order, the latter stores in permuted order => we only need to define the permutation indices once
# In the former, we use these row indices when loading X
# For the latter, we use these row indices when storing Y
# FUSE_MUL -> multiply routed outputs by their respective weights
# topk_weights are in token order
# Only account for the case when X is in expert order and we are permuting Y when fusing mul -- this precondition is checked in the interface
@triton.jit
def _grouped_gemm_forward_kernel(
x_ptr,
w_ptr,
y_ptr,
# Variable depending on routed probs
m_sizes_ptr,
gather_indices_ptr,
topk_weights_ptr,
# Constant problem shapes
NUM_EXPERTS: tl.constexpr,
NUM_TOKENS: tl.constexpr,
TOPK: tl.constexpr,
N: tl.constexpr,
K: tl.constexpr,
NUM_SMS: tl.constexpr,
# Tuning params
BLOCK_SIZE_M: tl.constexpr,
BLOCK_SIZE_N: tl.constexpr,
BLOCK_SIZE_K: tl.constexpr,
PERMUTE_X: tl.constexpr = False,
PERMUTE_Y: tl.constexpr = False,
FUSE_MUL_PRE: tl.constexpr = False,
FUSE_MUL_POST: tl.constexpr = False,
USE_FAST_ACCUM: tl.constexpr = False,
USE_TMA_LOAD_W: tl.constexpr = False,
USE_TMA_LOAD_X: tl.constexpr = False,
USE_TMA_STORE: tl.constexpr = False,
acc_dtype: tl.constexpr = tl.float32,
FLATTEN: tl.constexpr = True,
) -> None:
tl.static_assert(K % BLOCK_SIZE_K == 0)
TOTAL_TOKENS: tl.constexpr = NUM_TOKENS * TOPK
SHOULD_PERMUTE: tl.constexpr = PERMUTE_X or PERMUTE_Y
SHOULD_FUSE_MUL: tl.constexpr = FUSE_MUL_PRE or FUSE_MUL_POST
SHOULD_PERMUTE_OR_FUSE: tl.constexpr = SHOULD_PERMUTE or SHOULD_FUSE_MUL
# tl.static_print("SHOULD_PERMUTE", PERMUTE_X, PERMUTE_Y, FUSE_MUL_PRE, FUSE_MUL_POST, SHOULD_PERMUTE, SHOULD_FUSE, SHOULD_PERMUTE_OR_FUSE)
tidx = tl.program_id(0)
output_dtype: tl.dtype = y_ptr.dtype.element_ty
# Create TMA descriptors for loading sorted tokens
# When using TMA load, we don't permute_x, so shape should be [TOTAL_TOKENS, K]
# Also, we are defining a single global descriptor with single block shape
# Need to check that this does not result in errors when crossing expert boundaries
if USE_TMA_LOAD_X:
x_desc = tl._experimental_make_tensor_descriptor(
x_ptr,
shape=[TOTAL_TOKENS, K],
strides=[K, 1],
block_shape=[BLOCK_SIZE_M, BLOCK_SIZE_K],
)
if USE_TMA_LOAD_W:
expert_stride = N * K
w_desc = tl._experimental_make_tensor_descriptor(
w_ptr,
shape=[NUM_EXPERTS, N, K],
strides=[expert_stride, K, 1],
block_shape=[1, BLOCK_SIZE_N, BLOCK_SIZE_K],
)
m_end = 0
processed_tiles = 0
m_block_range = tl.arange(0, BLOCK_SIZE_M)
for expert_idx in tl.range(NUM_EXPERTS, flatten=FLATTEN):
m_start = m_end
m_size = tl.load(m_sizes_ptr + expert_idx).to(tl.int32)
m_end = m_start + m_size
if m_size > 0:
n_start = expert_idx * N
num_m_tiles = tl.cdiv(m_size, BLOCK_SIZE_M)
num_n_tiles = tl.cdiv(N, BLOCK_SIZE_N)
num_tiles_per_expert = num_m_tiles * num_n_tiles
# Need to create tma_store within loop since we need to predicate stores based on m_size
if USE_TMA_STORE:
y_desc = tl._experimental_make_tensor_descriptor(
y_ptr, # + m_start * N,
shape=[m_end, N],
strides=[N, 1],
block_shape=[BLOCK_SIZE_M, BLOCK_SIZE_N],
)
# Process tiles for this expert
while (
tidx >= processed_tiles
and tidx < processed_tiles + num_tiles_per_expert
):
tile_idx = tidx - processed_tiles
# Check if L2 cache re-use for this order is optimal
tile_m_idx = tile_idx % num_m_tiles
tile_n_idx = tile_idx // num_m_tiles
if SHOULD_PERMUTE_OR_FUSE:
# These will be used for loading and storing in permuted order
gather_offsets = tile_m_idx * BLOCK_SIZE_M + m_block_range
indices_to_gather = m_start + tl.max_contiguous(
tl.multiple_of(gather_offsets % m_size, BLOCK_SIZE_M),
BLOCK_SIZE_M,
)
expert_token_idx = tl.load(
gather_indices_ptr + indices_to_gather,
mask=indices_to_gather < TOTAL_TOKENS,
)
expert_token_offsets = expert_token_idx[:, None]
# Masks for permuted load and store
row_mask = gather_offsets < m_size
row_mask = row_mask[:, None]
# row_mask = indices_to_gather < m_end
# row_mask = row_mask[:, None]
# We only take into account the following two cases: (PERMUTE_X and NOT PERMUTE_Y) and (NOT PERMUTE_X and PERMUTE_Y)
# Hence, we can make the following simplifying assumptions when loading and storing
# Note the different strides between the two cases: the offsets for loading and storing are flipped and the strides must also be adjusted
if PERMUTE_X:
load_idx = (
(expert_token_offsets // TOPK) * K
) # Permute on load from token -> expert order, divide by TOPK to index from original number of tokens
store_idx = (
indices_to_gather[:, None] * N
) # Store in contiguous order
else:
off_am = tile_m_idx * BLOCK_SIZE_M
if not PERMUTE_Y:
# These will already be computed if permuting y
offs_am = off_am + m_block_range
row_mask = offs_am[:, None] < m_size
row_idx = m_start + offs_am[:, None]
store_idx = row_idx * N
if not USE_TMA_LOAD_X:
load_idx = row_idx * K
if PERMUTE_Y:
if not USE_TMA_LOAD_X:
load_idx = (
indices_to_gather[:, None] * K
) # Load in contiguous order (no permutation on load)
# offs_am = off_am + m_block_range
# row_mask = offs_am[:, None] < m_size
store_idx = (
expert_token_offsets * N
) # Permute on store from expert -> token order
# We always load topk weights in expert order
# In the pre-multiplication case, we multiply permuted hidden states by weights before the first gemm
# In the post-multiplication case, we multiply permuted hidden states by weights after the second gemm
# In either case, the hidden states are grouped by expert, so we always permute on load of topk weights
if SHOULD_FUSE_MUL:
topk_load_idx = expert_token_offsets
accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=acc_dtype)
offs_k = tl.arange(0, BLOCK_SIZE_K)
if not USE_TMA_LOAD_X:
x_ptrs = x_ptr + load_idx + offs_k[None, :]
if not USE_TMA_LOAD_W:
offs_bn = tile_n_idx * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
offs_bn = tl.max_contiguous(
tl.multiple_of(offs_bn % N, BLOCK_SIZE_N), BLOCK_SIZE_N
)
w_ptrs = w_ptr + (n_start + offs_bn[:, None]) * K + offs_k[None, :]
for k_offset in range(0, K, BLOCK_SIZE_K):
if not USE_TMA_LOAD_X:
x = tl.load(x_ptrs, mask=row_mask)
else:
x = x_desc.load([m_start + off_am, k_offset])
if FUSE_MUL_PRE:
# Check for correct broadcasting
topk_weights = tl.load(
topk_weights_ptr + topk_load_idx, mask=row_mask
)
x *= topk_weights.to(x.dtype)
if not USE_TMA_LOAD_W:
w = tl.load(w_ptrs, mask=offs_bn[:, None] < N)
else:
w = w_desc.load(
[expert_idx, tile_n_idx * BLOCK_SIZE_N, k_offset]
)
w = tl.reshape(w, (BLOCK_SIZE_N, BLOCK_SIZE_K))
accumulator += tl.dot(x, w.T)
if not USE_TMA_LOAD_X:
x_ptrs += BLOCK_SIZE_K
if not USE_TMA_LOAD_W:
w_ptrs += BLOCK_SIZE_K
y = accumulator.to(output_dtype)
# NOTE: order of fusing multiplication is important
# Fusing before accumulator dtype conversion results in numerical diffs
if FUSE_MUL_POST:
# Check for correct broadcasting
topk_weights = tl.load(
topk_weights_ptr + topk_load_idx, mask=row_mask
)
y *= topk_weights.to(output_dtype)
offs_bn = tile_n_idx * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
store_mask = row_mask & (offs_bn[None, :] < N)
if USE_TMA_STORE:
offset_m = tile_m_idx * BLOCK_SIZE_M # .to(tl.int32)
offset_n = tile_n_idx * BLOCK_SIZE_N # .to(tl.int32)
y_desc.store([m_start + offset_m, offset_n], y)
else:
tl.store(
y_ptr + store_idx + offs_bn[None, :],
y,
mask=store_mask,
)
tidx += NUM_SMS
processed_tiles += num_tiles_per_expert
_autotuned_grouped_gemm_forward_kernel = triton.autotune(
configs=get_forward_configs(),
prune_configs_by={"early_config_prune": prune_kernel_configs_fwd},
key=[
"NUM_EXPERTS",
"NUM_TOKENS",
"N",
"K",
"PERMUTE_X",
"PERMUTE_Y",
"FUSE_MUL_POST",
],
)(_grouped_gemm_forward_kernel)

View file

@ -0,0 +1,274 @@
"""
Manual tuning utils
"""
from collections import OrderedDict
from dataclasses import asdict, dataclass, fields
from itertools import product
from typing import Optional
import pandas as pd
import torch
import triton
from triton.runtime.errors import OutOfResources
from grouped_gemm.kernels.autotuning import (
BOOLS,
DEFAULT_K_BLOCK_SIZES,
DEFAULT_M_BLOCK_SIZES,
DEFAULT_N_BLOCK_SIZES,
DEFAULT_NUM_STAGES,
DEFAULT_NUM_WARPS,
)
@dataclass
class DeviceProperties:
NUM_SM: int
NUM_REGS: int
SIZE_SMEM: int
WARP_SIZE: int
_DEVICE_PROPERTIES: Optional[DeviceProperties] = None
def get_device_properties():
global _DEVICE_PROPERTIES
if _DEVICE_PROPERTIES is None:
properties = triton.runtime.driver.active.utils.get_device_properties(
torch.cuda.current_device()
)
NUM_SM = properties["multiprocessor_count"]
NUM_REGS = properties["max_num_regs"]
SIZE_SMEM = properties["max_shared_mem"]
WARP_SIZE = properties["warpSize"]
_DEVICE_PROPERTIES = DeviceProperties(NUM_SM, NUM_REGS, SIZE_SMEM, WARP_SIZE)
return _DEVICE_PROPERTIES
@dataclass
class KernelConfig:
BLOCK_SIZE_M: int = 32
BLOCK_SIZE_N: int = 32
BLOCK_SIZE_K: int = 32
num_warps: int = 4
num_stages: int = 2
flatten: bool = True
permute_x: bool = False
permute_y: bool = False
fuse_mul_post: bool = False
use_tma_store: bool = False
def to_string(self, include_tuning_params: bool = False, include_tma: bool = False):
s = []
if self.permute_x:
s.append("permute_x")
if self.permute_y:
s.append("permute_y")
if include_tuning_params:
s.append(
f"BLOCK_SIZE_M={self.BLOCK_SIZE_M},BLOCK_SIZE_N={self.BLOCK_SIZE_N},BLOCK_SIZE_K={self.BLOCK_SIZE_K},num_warps={self.num_warps},num_stages={self.num_stages},flatten={self.flatten}"
)
if include_tma:
for f in fields(self):
if f.name.startswith("use_tma_"):
if getattr(self, f.name):
s.append(f.name)
return ",".join(s)
@dataclass
class KernelConfigForward(KernelConfig):
use_tma_load_w: bool = False
use_tma_load_x: bool = False
@dataclass
class KernelConfigBackward_dW(KernelConfig):
use_tma_load_dy: bool = False
use_tma_load_x: bool = False
@dataclass
class KernelConfigBackward_dX(KernelConfig):
use_tma_load_dy: bool = False
use_tma_load_w: bool = False
@dataclass
class KernelResult:
torch_time: float
triton_time: float
speedup: float
kernel_config: KernelConfig
def to_dict(self):
return OrderedDict(
**asdict(self.kernel_config),
torch_time=self.torch_time,
triton_time=self.triton_time,
speedup=self.speedup,
)
@staticmethod
def to_dataframe(
results: list["KernelResult"], sort_by: str = "speedup", ascending: bool = False
):
df = pd.DataFrame([result.to_dict() for result in results])
df = df.sort_values(by=sort_by, ascending=ascending)
return df
@staticmethod
def to_csv(
results: list["KernelResult"],
sort_by: str = "speedup",
ascending: bool = False,
filename: str = "results.csv",
):
df = KernelResult.to_dataframe(results, sort_by, ascending)
df.to_csv(filename, index=False)
@staticmethod
def print_table(
results: list["KernelResult"],
sort_by: str = "speedup",
ascending: bool = False,
num_results: int = 10,
):
df = KernelResult.to_dataframe(results, sort_by, ascending)
print(df.head(num_results).to_string(index=False))
def get_kernel_configs(
BLOCK_M=DEFAULT_M_BLOCK_SIZES,
BLOCK_N=DEFAULT_N_BLOCK_SIZES,
BLOCK_K=DEFAULT_K_BLOCK_SIZES,
num_warps=DEFAULT_NUM_WARPS,
num_stages=DEFAULT_NUM_STAGES,
use_tma_loads=BOOLS,
fuse_permute=BOOLS,
):
kernel_configs_fwd = []
kernel_configs_backward_dW = []
kernel_configs_backward_dX = []
for block_m, block_n, block_k, w, s, use_tma_load, permute in product(
BLOCK_M, BLOCK_N, BLOCK_K, num_warps, num_stages, use_tma_loads, fuse_permute
):
kernel_configs_fwd.append(
KernelConfigForward(
BLOCK_SIZE_M=block_m,
BLOCK_SIZE_N=block_n,
BLOCK_SIZE_K=block_k,
num_warps=w,
num_stages=s,
use_tma_load_x=use_tma_load,
use_tma_load_w=use_tma_load,
use_tma_store=False,
permute_x=permute,
permute_y=permute,
)
)
kernel_configs_backward_dW.append(
KernelConfigBackward_dW(
BLOCK_SIZE_M=block_m,
BLOCK_SIZE_N=block_n,
BLOCK_SIZE_K=block_k,
num_warps=w,
num_stages=s,
use_tma_load_dy=use_tma_load,
use_tma_load_x=use_tma_load,
use_tma_store=False,
permute_x=permute,
permute_y=permute,
)
)
kernel_configs_backward_dX.append(
KernelConfigBackward_dX(
BLOCK_SIZE_M=block_m,
BLOCK_SIZE_N=block_n,
BLOCK_SIZE_K=block_k,
num_warps=w,
num_stages=s,
use_tma_load_dy=use_tma_load,
use_tma_load_w=use_tma_load,
use_tma_store=False,
permute_x=permute,
permute_y=permute,
)
)
kernel_configs_fwd = prune_kernel_configs_fwd(kernel_configs_fwd)
kernel_configs_backward_dW = prune_kernel_configs_backward_dW(
kernel_configs_backward_dW
)
kernel_configs_backward_dX = prune_kernel_configs_backward_dX(
kernel_configs_backward_dX
)
return kernel_configs_fwd, kernel_configs_backward_dW, kernel_configs_backward_dX
def prune_kernel_configs_fwd(configs: list[KernelConfigForward]):
pruned_configs = []
for config in configs:
if config.use_tma_load_x and config.permute_x:
continue
if config.permute_x and config.permute_y:
continue
if config.use_tma_store and config.permute_y:
continue
pruned_configs.append(config)
return pruned_configs
def prune_kernel_configs_backward_dX(configs: list[KernelConfigBackward_dX]):
pruned_configs = []
for config in configs:
if config.use_tma_load_dy and config.permute_y:
continue
if config.permute_x and config.permute_y:
continue
if config.use_tma_store and config.permute_x:
continue
pruned_configs.append(config)
return pruned_configs
def prune_kernel_configs_backward_dW(configs: list[KernelConfigBackward_dW]):
pruned_configs = []
for config in configs:
if config.use_tma_load_dy and config.permute_y:
continue
if config.use_tma_load_x and config.permute_x:
continue
if config.permute_x and config.permute_y:
continue
pruned_configs.append(config)
return pruned_configs
class TritonTuningContext:
def __init__(self, kernel_config: KernelConfig):
self.kernel_config = kernel_config
self.success = True
def __enter__(self):
# Setup code can be added here if needed
return self
def __exit__(self, exc_type, exc_value, traceback):
if exc_type is OutOfResources:
name = exc_value.name
required = exc_value.required
limit = exc_value.limit
print(
f"Kernel config {self.kernel_config} failed: {name}, required: {required}, limit: {limit}"
)
self.success = False
elif exc_type is not None:
print(
f"Error running Triton grouped GEMM for kernel config: {self.kernel_config}: {exc_value}"
)
self.success = False
# Return False to propagate exceptions, True to suppress them
return True

View file

@ -0,0 +1,158 @@
import torch
from transformers.models.qwen3_moe.configuration_qwen3_moe import Qwen3MoeConfig
from transformers.models.qwen3_moe.modeling_qwen3_moe import Qwen3MoeSparseMoeBlock
from grouped_gemm.interface import grouped_gemm
from grouped_gemm.kernels.tuning import (
KernelConfigBackward_dW,
KernelConfigBackward_dX,
KernelConfigForward,
)
from grouped_gemm.reference.moe_ops import (
Qwen3MoeGroupedGEMMBlock,
permute,
unpermute,
)
"""
Reference implementation of MoE block using grouped gemm.
This is the same as the Qwen3MoeGroupedGEMMBlock but with triton grouped gemm in place of torch-native grouped gemm implementation.
NOTE: This is NOT to be used for production as it contains many extra checks and saves all intermediate results for debugging.
"""
class Qwen3MoeFusedGroupedGEMMBlock(Qwen3MoeGroupedGEMMBlock):
def __init__(
self,
config: Qwen3MoeConfig,
gate: torch.Tensor,
gate_up_proj: torch.Tensor,
down_proj: torch.Tensor,
permute_x: bool = True,
permute_y: bool = True,
autotune: bool = True,
kernel_config_fwd: KernelConfigForward = None,
kernel_config_bwd_dW: KernelConfigBackward_dW = None,
kernel_config_bwd_dX: KernelConfigBackward_dX = None,
dW_only: bool = False,
dX_only: bool = False,
):
super().__init__(config, gate, gate_up_proj, down_proj)
self.permute_x = permute_x
self.permute_y = permute_y
self.autotune = autotune
if not autotune:
assert (
kernel_config_fwd is not None
and kernel_config_bwd_dW is not None
and kernel_config_bwd_dX is not None
), "Kernel configs must be provided if autotune is False"
self.kernel_config_fwd = kernel_config_fwd
self.kernel_config_bwd_dW = kernel_config_bwd_dW
self.kernel_config_bwd_dX = kernel_config_bwd_dX
self.dW_only = dW_only
self.dX_only = dX_only
@classmethod
def from_hf(
cls,
moe_block: Qwen3MoeSparseMoeBlock,
permute_x: bool = True,
permute_y: bool = True,
autotune: bool = True,
kernel_config_fwd: KernelConfigForward = None,
kernel_config_bwd_dW: KernelConfigBackward_dW = None,
kernel_config_bwd_dX: KernelConfigBackward_dX = None,
dW_only: bool = False,
dX_only: bool = False,
):
config: Qwen3MoeConfig = moe_block.experts[0].config
gate, gate_up_proj, down_proj = Qwen3MoeGroupedGEMMBlock.extract_hf_weights(
moe_block
)
return cls(
config,
gate,
gate_up_proj,
down_proj,
permute_x=permute_x,
permute_y=permute_y,
autotune=autotune,
kernel_config_fwd=kernel_config_fwd,
kernel_config_bwd_dW=kernel_config_bwd_dW,
kernel_config_bwd_dX=kernel_config_bwd_dX,
dW_only=dW_only,
dX_only=dX_only,
)
def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
batch_size, sequence_length, hidden_dim = hidden_states.shape
num_tokens = batch_size * sequence_length
total_tokens = num_tokens * self.top_k
hidden_states = hidden_states.view(-1, hidden_dim)
router_logits, routing_weights, selected_experts = self.run_router(
hidden_states
)
# Pre-processing
# 1. Compute tokens per expert and indices for gathering tokes from token order to expert order
# NOTE: these are auxiliary data structs which don't need to be recorded in autograd graph
token_counts_by_expert, gather_indices = (
self.get_token_counts_and_gather_indices(selected_experts)
)
# 2. permute_x -> permutation will be fused in prologue of first grouped gemm
if not self.permute_x:
hidden_states = permute(hidden_states, gather_indices, self.top_k)
# Start expert computation
hidden_states = grouped_gemm(
X=hidden_states,
W=self.gate_up_proj,
m_sizes=token_counts_by_expert,
gather_indices=gather_indices,
topk=self.top_k,
permute_x=self.permute_x,
permute_y=False, # output of first grouped gemm should never be permuted
autotune=self.autotune,
kernel_config_fwd=self.kernel_config_fwd,
kernel_config_bwd_dW=self.kernel_config_bwd_dW,
kernel_config_bwd_dX=self.kernel_config_bwd_dX,
is_first_gemm=True,
dW_only=self.dW_only,
dX_only=self.dX_only,
)
hidden_states = self.act_and_mul(hidden_states)
hidden_states = grouped_gemm(
X=hidden_states,
W=self.down_proj,
m_sizes=token_counts_by_expert,
gather_indices=gather_indices,
topk=self.top_k,
permute_x=False,
permute_y=self.permute_y,
autotune=self.autotune,
kernel_config_fwd=self.kernel_config_fwd,
kernel_config_bwd_dW=self.kernel_config_bwd_dW,
kernel_config_bwd_dX=self.kernel_config_bwd_dX,
is_first_gemm=False,
dW_only=self.dW_only,
dX_only=self.dX_only,
)
# Post-processing
# 1. Unpermute from expert order to token order
if not self.permute_y:
hidden_states = unpermute(hidden_states, gather_indices)
# 2. Merge topk weights
hidden_states = (
hidden_states.view(num_tokens, self.top_k, hidden_dim)
* routing_weights[..., None]
)
hidden_states = hidden_states.sum(dim=1)
hidden_states = hidden_states.view(batch_size, sequence_length, hidden_dim)
return hidden_states, router_logits

View file

@ -0,0 +1,335 @@
from dataclasses import dataclass
from typing import Tuple
import torch
import torch.nn as nn
import torch.nn.functional as F
from transformers.models.qwen3_moe import Qwen3MoeConfig
from transformers.models.qwen3_moe.modeling_qwen3_moe import (
ACT2FN,
Qwen3MoeSparseMoeBlock,
)
def permute(X: torch.Tensor, gather_indices: torch.Tensor, topk: int):
"""
Scatters X to a new tensor with shape [total_tokens, hidden_dim] where total_tokens is num_tokens * topk,
permuting the tokens according to sorted_token_idx.
Helper for grouped gemm where hidden states need be ordered by expert.
X: [num_tokens, hidden_dim]
sorted_token_idx: [num_tokens * topk]
topk: int
Returns:
[total_tokens, hidden_dim]
"""
assert gather_indices.ndim == 1
X = X.view(-1, X.shape[-1])
# Shortcut for topk == 1
if topk == 1:
return X[gather_indices]
return X[gather_indices // topk]
def unpermute(X: torch.Tensor, gather_indices: torch.Tensor):
X = X.view(-1, X.shape[-1]) if X.ndim > 2 else X
unpermuted = torch.empty_like(X)
unpermuted.index_copy_(0, gather_indices, X)
return unpermuted.view_as(X)
def calculate_topk(
gating_output: torch.Tensor,
top_k: int,
use_sigmoid: bool,
renormalize: bool,
pre_act: bool = True,
post_act: bool = False,
):
"""
If post_act is True, then activation function is run AFTER topk
If post_act is False, then activation function is run BEFORE topk
This is to align with triton_bench implementation (post_act) whereas most models use pre_act (e.g. llama4, deepseek)
"""
assert pre_act ^ post_act, "only one of pre_act or post_act can be True"
def _activation(gating_output: torch.Tensor):
if use_sigmoid:
scores = torch.sigmoid(gating_output.to(torch.float32)).to(
gating_output.dtype
)
else:
scores = F.softmax(gating_output.to(torch.float32), dim=1).to(
gating_output.dtype
)
return scores
if pre_act:
scores = _activation(gating_output)
else:
scores = gating_output
topk_weights, topk_ids = torch.topk(scores, k=top_k, dim=1)
if post_act:
topk_weights = _activation(topk_weights)
if renormalize:
topk_weights /= torch.sum(topk_weights, dim=-1, keepdim=True).to(
gating_output.dtype
)
return topk_weights, topk_ids
@torch.no_grad()
def get_routing_indices(
selected_experts, num_experts, return_scatter_indices: bool = False
):
"""
Returns:
token_counts_by_expert: [num_experts]
gather_indices: [num_tokens]
scatter_indices [Optional] (torch.Tensor):
Indices for unpermuting gathered inputs back to token order, shape ``(bs * seqlen * top_k,)``.
"""
# group tokens together by expert indices from 0 to num_experts and pass that to experts forward
token_counts_by_expert = torch.histc(
selected_experts.view(-1),
bins=num_experts,
min=0,
max=num_experts,
)
# token_indices_experts_sorted shape (bs*slen*top_k,)
gather_indices = torch.argsort(selected_experts.view(-1), stable=True)
if return_scatter_indices:
scatter_indices = gather_indices.argsort()
return token_counts_by_expert, gather_indices, scatter_indices
else:
return token_counts_by_expert, gather_indices
def torch_grouped_gemm(X, W, m_sizes, transpose=True):
"""
X: [M, K] if forward, else [M, N]
W: [E, N, K]
m_sizes: [E]
Returns:
Y: [M, N] if forward, else [M, K]
"""
X = X.view(-1, X.shape[-1])
M, K = X.shape
assert m_sizes.ndim == 1
E = m_sizes.shape[0]
assert W.ndim == 3
assert W.shape[0] == E
N = W.shape[1]
result = torch.zeros((M, N), dtype=X.dtype, device=X.device)
m_start = 0
for g in range(E):
m_size = m_sizes[g]
if m_size > 0:
m_end = m_start + m_size
# Extract group input
# m_size x K
X_g = X[m_start:m_end]
# N x K
W_g = W[g]
# Y_g = X_g @ W_g.T -> [m_size, N]
W_g = W_g.T if transpose else W_g
Y_g = X_g @ W_g
result[m_start:m_end] = Y_g
m_start = m_end
return result
@dataclass
class GroupedGEMMResult:
token_counts_by_expert: torch.Tensor
gather_indices: torch.Tensor
topk_weights: torch.Tensor
first_gemm: torch.Tensor
intermediate: torch.Tensor
second_gemm: torch.Tensor
hidden_states_unpermute: torch.Tensor
hidden_states: torch.Tensor # final output
class Qwen3MoeGroupedGEMMBlock(nn.Module):
def __init__(
self,
config,
gate: torch.Tensor,
gate_up_proj: torch.Tensor,
down_proj: torch.Tensor,
):
super().__init__()
self.num_experts = config.num_experts
self.top_k = config.num_experts_per_tok
self.norm_topk_prob = config.norm_topk_prob
self.hidden_size = config.hidden_size
self.moe_intermediate_size = config.moe_intermediate_size
assert gate.shape == (config.num_experts, config.hidden_size)
assert gate_up_proj.shape == (
config.num_experts,
2 * config.moe_intermediate_size,
config.hidden_size,
)
assert down_proj.shape == (
config.num_experts,
config.hidden_size,
config.moe_intermediate_size,
)
# gating
self.gate = torch.nn.Parameter(gate)
# experts
self.gate_up_proj = torch.nn.Parameter(gate_up_proj, requires_grad=True)
self.down_proj = torch.nn.Parameter(down_proj, requires_grad=True)
self.act_fn = ACT2FN[config.hidden_act]
@staticmethod
def extract_hf_weights(moe_block: Qwen3MoeSparseMoeBlock):
config: Qwen3MoeConfig = moe_block.experts[0].config
num_experts = config.num_experts
gate = moe_block.gate.weight.data
gate_proj = torch.stack(
[moe_block.experts[i].gate_proj.weight.data for i in range(num_experts)],
dim=0,
)
up_proj = torch.stack(
[moe_block.experts[i].up_proj.weight.data for i in range(num_experts)],
dim=0,
)
down_proj = torch.stack(
[moe_block.experts[i].down_proj.weight.data for i in range(num_experts)],
dim=0,
)
gate_up_proj = torch.cat([gate_proj, up_proj], dim=1)
return gate, gate_up_proj, down_proj
@classmethod
def from_hf(cls, moe_block: Qwen3MoeSparseMoeBlock):
config: Qwen3MoeConfig = moe_block.experts[0].config
gate, gate_up_proj, down_proj = cls.extract_hf_weights(moe_block)
return cls(config, gate, gate_up_proj, down_proj)
def check_weights(self, moe_block: Qwen3MoeSparseMoeBlock):
for i in range(self.num_experts):
assert self.gate_up_proj[i].equal(
torch.cat(
[
moe_block.experts[i].gate_proj.weight.data,
moe_block.experts[i].up_proj.weight.data,
],
dim=0,
)
)
assert self.down_proj[i].equal(moe_block.experts[i].down_proj.weight.data)
def act_and_mul(self, x: torch.Tensor) -> torch.Tensor:
assert x.shape[-1] == 2 * self.moe_intermediate_size
gate_proj = x[..., : self.moe_intermediate_size]
up_proj = x[..., self.moe_intermediate_size :]
return self.act_fn(gate_proj) * up_proj
def run_router(self, hidden_states: torch.Tensor) -> torch.Tensor:
# router_logits: (batch * sequence_length, n_experts)
router_logits = torch.nn.functional.linear(hidden_states, self.gate)
routing_weights = F.softmax(router_logits, dim=1, dtype=torch.float)
routing_weights, selected_experts = torch.topk(
routing_weights, self.top_k, dim=-1
)
if self.norm_topk_prob: # only diff with mixtral sparse moe block!
routing_weights /= routing_weights.sum(dim=-1, keepdim=True)
# we cast back to the input dtype
routing_weights = routing_weights.to(hidden_states.dtype)
return router_logits, routing_weights, selected_experts
def get_token_counts_and_gather_indices(
self, selected_experts: torch.Tensor
) -> Tuple[torch.Tensor, torch.Tensor]:
token_counts_by_expert, gather_indices = get_routing_indices(
selected_experts, self.num_experts
)
assert not token_counts_by_expert.requires_grad
assert not gather_indices.requires_grad
return token_counts_by_expert, gather_indices
def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
""" """
batch_size, sequence_length, hidden_dim = hidden_states.shape
num_tokens = batch_size * sequence_length
total_tokens = num_tokens * self.top_k
hidden_states = hidden_states.view(-1, hidden_dim)
router_logits, routing_weights, selected_experts = self.run_router(
hidden_states
)
# 1. Compute tokens per expert and indices for gathering tokes from token order to expert order
# NOTE: these are auxiliary data structs which don't need to be recorded in autograd graph
token_counts_by_expert, gather_indices = (
self.get_token_counts_and_gather_indices(selected_experts)
)
# 2. Permute tokens from token order to expert order
hidden_states = permute(hidden_states, gather_indices, self.top_k)
assert hidden_states.shape == (total_tokens, hidden_dim)
# Start expert computation
first_gemm = torch_grouped_gemm(
X=hidden_states, W=self.gate_up_proj, m_sizes=token_counts_by_expert
)
assert first_gemm.shape == (total_tokens, 2 * self.moe_intermediate_size)
intermediate = self.act_and_mul(first_gemm)
assert intermediate.shape == (total_tokens, self.moe_intermediate_size)
second_gemm = torch_grouped_gemm(
X=intermediate, W=self.down_proj, m_sizes=token_counts_by_expert
)
assert second_gemm.shape == (total_tokens, hidden_dim)
# Post-processing
# 1. Unpermute from expert order to token order
hidden_states_unpermute = unpermute(second_gemm, gather_indices)
assert hidden_states_unpermute.shape == (total_tokens, hidden_dim)
# 2. Merge topk weights
hidden_states = (
hidden_states_unpermute.view(num_tokens, self.top_k, hidden_dim)
* routing_weights[..., None]
)
hidden_states = hidden_states.sum(dim=1)
assert hidden_states.shape == (num_tokens, hidden_dim)
hidden_states = hidden_states.view(batch_size, sequence_length, hidden_dim)
return GroupedGEMMResult(
token_counts_by_expert=token_counts_by_expert,
gather_indices=gather_indices,
topk_weights=routing_weights,
first_gemm=first_gemm,
intermediate=intermediate,
second_gemm=second_gemm,
hidden_states_unpermute=hidden_states_unpermute,
hidden_states=hidden_states,
), router_logits

View file

@ -0,0 +1,5 @@
torch
git+https://github.com/huggingface/transformers.git@main
pytest
pandas
ruff

View file

View file

@ -0,0 +1,333 @@
import itertools
from contextlib import contextmanager
from dataclasses import dataclass, field
import torch
from grouped_gemm.kernels.tuning import (
KernelConfig,
KernelConfigBackward_dW,
KernelConfigBackward_dX,
KernelConfigForward,
prune_kernel_configs_backward_dW,
prune_kernel_configs_backward_dX,
prune_kernel_configs_fwd,
)
def print_delimiter(char="-", length=80):
print(char * length)
@contextmanager
def delimiter_context():
print_delimiter()
yield
print_delimiter()
def make_inputs(M, N, K, E, topk, dtype, requires_grad=False):
X1 = (
torch.randn((M, K), device="cuda", dtype=dtype, requires_grad=requires_grad)
/ 10
)
X2 = (
torch.randn(
(M * topk, N), device="cuda", dtype=dtype, requires_grad=requires_grad
)
/ 10
)
W1 = (
torch.randn(
(E, 2 * N, K), device="cuda", dtype=dtype, requires_grad=requires_grad
)
/ 10
)
W2 = (
torch.randn((E, K, N), device="cuda", dtype=dtype, requires_grad=requires_grad)
/ 10
)
score = torch.randn((M, E), device="cuda", dtype=dtype, requires_grad=requires_grad)
if requires_grad:
X1.retain_grad()
X2.retain_grad()
W1.retain_grad()
W2.retain_grad()
score.retain_grad()
return X1, X2, W1, W2, score
@dataclass(kw_only=True)
class DataConfig:
seq_len: int
dtype: torch.dtype
device: str = "cuda"
bs: int = 1
@dataclass(kw_only=True)
class ModelConfig:
hidden_size: int
intermediate_size: int
num_experts: int
topk: int
use_sigmoid: bool
renormalize: bool
pre_mul: bool = False
post_mul: bool = field(init=False)
def __post_init__(self):
self.post_mul = not self.pre_mul
@dataclass(kw_only=True)
class GroupedGEMMTestConfig:
name: str = "test"
data_config: DataConfig
model_config: ModelConfig
TOLERANCE = {
torch.bfloat16: (1e-3, 1e-3),
torch.float16: (1e-4, 1e-4),
torch.float32: (1e-5, 1e-5),
}
# from https://github.com/triton-lang/triton/blob/main/bench/triton_bench/testing.py
def assert_equal(ref, tri):
if isinstance(ref, torch.Tensor):
assert torch.all(ref == tri), f"tensors not equal {ref} != {tri}"
else:
assert ref == tri, f"ref not equal to tri {ref} != {tri}"
def assert_close(ref, tri, maxtol=None, rmstol=None, description="--", verbose=True):
if tri.dtype.itemsize == 1:
ref_as_type = ref.to(tri.dtype)
if ref.dtype == tri.dtype:
assert torch.all(ref_as_type == tri)
return
ref = ref_as_type
if maxtol is None:
maxtol = 2e-2
if rmstol is None:
rmstol = 4e-3
"""
Compare reference values against obtained values.
"""
# cast to float32:
ref = ref.to(torch.float32).detach()
tri = tri.to(torch.float32).detach()
assert ref.shape == tri.shape, (
f"Tensors must have same size {ref.shape=} {tri.shape=}"
)
# deal with infinite elements:
inf_mask_ref = torch.isinf(ref)
inf_mask_tri = torch.isinf(tri)
assert torch.equal(inf_mask_ref, inf_mask_tri), (
"Tensor must have same infinite elements"
)
refn = torch.where(inf_mask_ref, 0, ref)
trin = torch.where(inf_mask_tri, 0, tri)
# normalise so that RMS calculation doesn't overflow:
eps = 1.0e-30
multiplier = 1.0 / (torch.max(torch.abs(refn)) + eps)
refn *= multiplier
trin *= multiplier
ref_rms = torch.sqrt(torch.square(refn).mean()) + eps
rel_err = torch.abs(refn - trin) / torch.maximum(ref_rms, torch.abs(refn))
max_err = torch.max(rel_err).item()
rms_err = torch.sqrt(torch.square(rel_err).mean()).item()
if verbose:
print(
"%s maximum relative error = %s (threshold = %s)"
% (description, max_err, maxtol)
)
print(
"%s RMS relative error = %s (threshold = %s)"
% (description, rms_err, rmstol)
)
if max_err > maxtol:
bad_idxs = torch.nonzero(rel_err > maxtol)
num_nonzero = bad_idxs.size(0)
bad_idxs = bad_idxs[:1000]
print(
"%d / %d mismatched elements (shape = %s) at coords %s"
% (num_nonzero, rel_err.numel(), tuple(rel_err.shape), bad_idxs.tolist())
)
bad_idxs = bad_idxs.unbind(-1)
print("ref values: ", ref[*bad_idxs].cpu())
print("tri values: ", tri[*bad_idxs].cpu())
assert max_err <= maxtol
assert rms_err <= rmstol
def assert_indx_equal(ref, tri):
assert_equal(ref, tri[: len(ref)])
assert torch.all(tri[len(ref) :] == -1)
def get_kernel_test_configs(
BLOCK_SIZE_M=32,
BLOCK_SIZE_N=32,
BLOCK_SIZE_K=32,
num_warps=4,
num_stages=2,
) -> list[KernelConfig]:
configs_fwd = []
configs_bwd_dX = []
configs_bwd_dW = []
for permute_x in [False, True]:
for permute_y in [False, True]:
for use_tma_load_w in [True, False]:
for use_tma_load_x in [True, False]:
for use_tma_store in [True, False]:
configs_fwd.append(
KernelConfigForward(
BLOCK_SIZE_M=BLOCK_SIZE_M,
BLOCK_SIZE_N=BLOCK_SIZE_N,
BLOCK_SIZE_K=BLOCK_SIZE_K,
num_warps=num_warps,
num_stages=num_stages,
use_tma_load_w=use_tma_load_w,
use_tma_load_x=use_tma_load_x,
use_tma_store=use_tma_store,
permute_x=permute_x,
permute_y=permute_y,
)
)
configs_bwd_dX.append(
KernelConfigBackward_dX(
BLOCK_SIZE_M=BLOCK_SIZE_M,
BLOCK_SIZE_N=BLOCK_SIZE_N,
BLOCK_SIZE_K=BLOCK_SIZE_K,
num_warps=num_warps,
num_stages=num_stages,
use_tma_load_dy=use_tma_load_x,
use_tma_load_w=use_tma_load_w,
permute_x=permute_x,
permute_y=permute_y,
use_tma_store=use_tma_store,
)
)
configs_bwd_dW.append(
KernelConfigBackward_dW(
BLOCK_SIZE_M=BLOCK_SIZE_M,
BLOCK_SIZE_N=BLOCK_SIZE_N,
BLOCK_SIZE_K=BLOCK_SIZE_K,
num_warps=num_warps,
num_stages=num_stages,
use_tma_load_dy=use_tma_load_w,
use_tma_load_x=use_tma_load_x,
permute_x=permute_x,
permute_y=permute_y,
use_tma_store=use_tma_store,
)
)
configs_fwd = prune_kernel_configs_fwd(configs_fwd)
configs_bwd_dX = prune_kernel_configs_backward_dX(configs_bwd_dX)
configs_bwd_dW = prune_kernel_configs_backward_dW(configs_bwd_dW)
return configs_fwd, configs_bwd_dX, configs_bwd_dW
def remove_feature_flags(
kernel_configs: list[KernelConfig],
permute_x: bool = True,
permute_y: bool = True,
tma_loads: bool = True,
tma_store: bool = True,
):
pruned_configs = []
for config in kernel_configs:
# Remove permute flags first:
if permute_x and config.permute_x:
continue
if permute_y and config.permute_y:
continue
if tma_loads:
if isinstance(config, KernelConfigForward):
if config.use_tma_load_w or config.use_tma_load_x:
continue
if isinstance(config, KernelConfigBackward_dX):
if config.use_tma_load_dy or config.use_tma_load_w:
continue
if isinstance(config, KernelConfigBackward_dW):
if config.use_tma_load_dy or config.use_tma_load_x:
continue
if tma_store:
if config.use_tma_store:
continue
pruned_configs.append(config)
return pruned_configs
# Test Configs
TOPK = [1, 4]
NUM_EXPERTS = [4, 16]
TEST_MODEL_SIZES = [
(32, 32), # Debug
(128, 128), # Small
(512, 512), # Medium
]
SMALL_MODEL_CONFIGS = [
ModelConfig(
topk=topk,
num_experts=num_experts,
hidden_size=model_size[0],
intermediate_size=model_size[1],
use_sigmoid=False,
renormalize=False,
)
for topk, num_experts, model_size in itertools.product(
TOPK, NUM_EXPERTS, TEST_MODEL_SIZES
)
]
LLAMA_MODEL_CONFIG = ModelConfig(
topk=1,
num_experts=16,
hidden_size=5120,
intermediate_size=8192,
use_sigmoid=True,
renormalize=False,
)
QWEN_MODEL_CONFIG = ModelConfig(
topk=8,
num_experts=128,
hidden_size=2048,
intermediate_size=768,
use_sigmoid=False,
renormalize=False,
)
SEQLENS = [128, 1024]
DTYPE = [torch.bfloat16]
DATA_CONFIGS = [
DataConfig(seq_len=seq_len, dtype=dtype)
for seq_len, dtype in itertools.product(SEQLENS, DTYPE)
]
KERNEL_CONFIGS_FWD, KERNEL_CONFIGS_BWD_dX, KERNEL_CONFIGS_BWD_dW = (
get_kernel_test_configs()
)
if __name__ == "__main__":
print(
KERNEL_CONFIGS_BWD_dX[0].to_string(
include_tuning_params=False, include_tma=False
)
)

View file

@ -0,0 +1,505 @@
from dataclasses import dataclass, fields
import torch
import torch.nn as nn
from huggingface_hub import HfApi
from huggingface_hub.utils import _safetensors
from transformers.models.qwen3_moe.configuration_qwen3_moe import Qwen3MoeConfig
from transformers.models.qwen3_moe.modeling_qwen3_moe import Qwen3MoeSparseMoeBlock
from grouped_gemm.interface import grouped_gemm
from grouped_gemm.kernels.tuning import (
KernelConfigBackward_dW,
KernelConfigBackward_dX,
KernelConfigForward,
)
from grouped_gemm.reference.moe_ops import (
GroupedGEMMResult,
Qwen3MoeGroupedGEMMBlock,
permute,
unpermute,
)
def rebind_experts_to_shared_buffer(
moe_block: Qwen3MoeSparseMoeBlock, config: Qwen3MoeConfig
):
num_experts = config.num_experts
hidden_size = config.hidden_size
interm_size = config.moe_intermediate_size
device = moe_block.experts[0].down_proj.weight.device
dtype = moe_block.experts[0].down_proj.weight.dtype
buffer_up = torch.empty(
num_experts, interm_size, hidden_size, device=device, dtype=dtype
)
buffer_gate = torch.empty(
num_experts, interm_size, hidden_size, device=device, dtype=dtype
)
buffer_down = torch.empty(
num_experts, hidden_size, interm_size, device=device, dtype=dtype
)
# Step 2: Copy existing expert weights into buffers
for i, expert in enumerate(moe_block.experts):
buffer_up[i].copy_(expert.up_proj.weight.data)
buffer_gate[i].copy_(expert.gate_proj.weight.data)
buffer_down[i].copy_(expert.down_proj.weight.data)
# Step 3: Rebind expert weights to views in shared buffer
for i, expert in enumerate(moe_block.experts):
expert.up_proj.weight = torch.nn.Parameter(buffer_up[i])
expert.gate_proj.weight = torch.nn.Parameter(buffer_gate[i])
expert.down_proj.weight = torch.nn.Parameter(buffer_down[i])
return buffer_up, buffer_gate, buffer_down
def get_expert_metadata(model_id: str):
api = HfApi()
metadata: _safetensors.SafetensorsRepoMetadata = api.get_safetensors_metadata(
model_id
)
return metadata.files_metadata
def clone_experts(
moe_block: Qwen3MoeSparseMoeBlock, config: Qwen3MoeConfig, copy: bool = True
):
down_projs = torch.empty(
config.num_experts, config.hidden_size, config.moe_intermediate_size
)
up_projs = torch.empty(
config.num_experts, config.moe_intermediate_size, config.hidden_size
)
gate_projs = torch.empty(
config.num_experts, config.moe_intermediate_size, config.hidden_size
)
for expert_idx, expert in enumerate(moe_block.experts):
down_projs[expert_idx].copy_(expert.down_proj.weight.data)
up_projs[expert_idx].copy_(expert.up_proj.weight.data)
gate_projs[expert_idx].copy_(expert.gate_proj.weight.data)
return gate_projs, up_projs, down_projs
@dataclass
class ForwardResult:
output: torch.Tensor
router_logits: torch.Tensor
X: torch.Tensor
# When using grouped gemm MoE implementation to additional debugging / checking of intermediate results
grouped_gemm_result: GroupedGEMMResult = None
@dataclass
class BackwardResult:
X_grad: torch.Tensor
gate_grad: torch.Tensor
gate_proj_grad: torch.Tensor
up_proj_grad: torch.Tensor
down_proj_grad: torch.Tensor
def check_down_proj_grad(
moe_block: Qwen3MoeSparseMoeBlock,
grouped_gemm_block: Qwen3MoeGroupedGEMMBlock,
atol: float,
rtol: float,
):
for i, expert in enumerate(moe_block.experts):
ref_grad = expert.down_proj.weight.grad
assert ref_grad is not None
test_grad = grouped_gemm_block.down_proj.grad[i]
assert test_grad is not None
diff = (ref_grad - test_grad).abs().max()
if not torch.allclose(ref_grad, test_grad, atol=atol, rtol=rtol):
print(f"expert {i} down_proj_grad_diff: {diff.detach().cpu().item():.6f}")
def check_gate_up_proj_grad(
moe_block: Qwen3MoeSparseMoeBlock,
grouped_gemm_block: Qwen3MoeGroupedGEMMBlock,
atol: float,
rtol: float,
):
moe_intermediate_size = grouped_gemm_block.moe_intermediate_size
for i, expert in enumerate(moe_block.experts):
ref_gate_proj_grad = expert.gate_proj.weight.grad
ref_up_proj_grad = expert.up_proj.weight.grad
assert ref_gate_proj_grad is not None
assert ref_up_proj_grad is not None
# Extract gradients
test_gate_proj_grad = grouped_gemm_block.gate_up_proj.grad[
i, :moe_intermediate_size
]
test_up_proj_grad = grouped_gemm_block.gate_up_proj.grad[
i, moe_intermediate_size:
]
assert test_gate_proj_grad is not None
assert test_up_proj_grad is not None
# Sanity check shapes
assert ref_gate_proj_grad.shape == test_gate_proj_grad.shape, (
f"{ref_gate_proj_grad.shape} != {test_gate_proj_grad.shape}"
)
assert ref_up_proj_grad.shape == test_up_proj_grad.shape, (
f"{ref_up_proj_grad.shape} != {test_up_proj_grad.shape}"
)
# Check gradients
diff = (ref_gate_proj_grad - test_gate_proj_grad).abs().max()
if not torch.allclose(
ref_gate_proj_grad, test_gate_proj_grad, atol=atol, rtol=rtol
):
print(f"expert {i} gate_proj_grad_diff: {diff.detach().cpu().item():.6f}")
diff = (ref_up_proj_grad - test_up_proj_grad).abs().max()
if not torch.allclose(
ref_up_proj_grad, test_up_proj_grad, atol=atol, rtol=rtol
):
print(f"expert {i} up_proj_grad_diff: {diff.detach().cpu().item():.6f}")
def check_gate_grad(
moe_block: Qwen3MoeSparseMoeBlock,
grouped_gemm_block: Qwen3MoeGroupedGEMMBlock,
atol: float,
rtol: float,
):
ref_grad = moe_block.gate.weight.grad
assert ref_grad is not None
test_grad = grouped_gemm_block.gate.grad
assert test_grad is not None
diff = (ref_grad - test_grad).abs().max()
if not torch.allclose(ref_grad, test_grad, atol=atol, rtol=rtol):
print(f"gate_grad_diff: {diff.detach().cpu().item():.6f}")
def check_wgrad(
moe_block: Qwen3MoeSparseMoeBlock,
grouped_gemm_block: Qwen3MoeGroupedGEMMBlock,
atol: float,
rtol: float,
):
check_down_proj_grad(moe_block, grouped_gemm_block, atol, rtol)
check_gate_up_proj_grad(moe_block, grouped_gemm_block, atol, rtol)
check_gate_grad(moe_block, grouped_gemm_block, atol, rtol)
def check_tensor_allclose(
X_ref: torch.Tensor,
X_test: torch.Tensor,
atol: float,
rtol: float,
name: str,
verbose: bool = False,
):
diff = (X_ref - X_test).abs().max()
if verbose:
print(f"{name} diff: {diff.detach().cpu().item():.6f}")
assert torch.allclose(X_ref, X_test, atol=atol, rtol=rtol), (
f"{name} diff: {diff.detach().cpu().item():.6f}"
)
def check_expert_grads(
ref_result: BackwardResult,
test_result: BackwardResult,
atol: float,
rtol: float,
verbose: bool = False,
):
fields_to_check = [f.name for f in fields(BackwardResult) if "proj" in f.name]
assert len(fields_to_check) == 3
for field in fields_to_check:
ref_grads = getattr(ref_result, field)
test_grads = getattr(test_result, field)
assert ref_grads.shape == test_grads.shape, (
f"{field}: {ref_grads.shape} != {test_grads.shape}"
)
# Test each expert
for i in range(ref_grads.shape[0]):
ref_grad = ref_grads[i]
test_grad = test_grads[i]
diff = (ref_grad - test_grad).abs().max()
assert torch.allclose(ref_grad, test_grad, atol=atol, rtol=rtol), (
f"{field}[{i}] diff: {diff.detach().cpu().item():.6f}"
)
# Test all experts
diff = (ref_grads - test_grads).abs().max()
if verbose:
print(f"{field} diff: {diff.detach().cpu().item():.6f}")
assert torch.allclose(ref_grads, test_grads, atol=atol, rtol=rtol), (
f"{field} diff: {diff.detach().cpu().item():.6f}"
)
def check_grads(
ref_result: BackwardResult,
test_result: BackwardResult,
atol: float,
rtol: float,
verbose: bool = False,
):
check_tensor_allclose(
ref_result.X_grad, test_result.X_grad, atol, rtol, "X.grad", verbose
)
check_tensor_allclose(
ref_result.gate_grad, test_result.gate_grad, atol, rtol, "gate.grad", verbose
)
check_expert_grads(ref_result, test_result, atol, rtol, verbose)
def check_fwd(
ref_result: ForwardResult,
test_result: ForwardResult,
atol: float,
rtol: float,
verbose: bool = False,
):
# First check hidden states (output)
ref_output = ref_result.output
test_output = test_result.output
diff = (ref_output - test_output).abs().max()
if verbose:
print(f"output diff: {diff.detach().cpu().item():.6f}")
assert torch.allclose(ref_output, test_output, atol=atol, rtol=rtol), (
f"output diff: {diff.detach().cpu().item():.6f}"
)
# Check router logits
ref_router_logits = ref_result.router_logits
test_router_logits = test_result.router_logits
diff = (ref_router_logits - test_router_logits).abs().max()
if verbose:
print(f"router_logits diff: {diff.detach().cpu().item():.6f}")
assert torch.allclose(
ref_router_logits, test_router_logits, atol=atol, rtol=rtol
), f"router_logits diff: {diff.detach().cpu().item():.6f}"
def check_grouped_gemm_results(
grouped_result: GroupedGEMMResult,
fused_result: GroupedGEMMResult,
permute_y: bool,
atol: float,
rtol: float,
verbose: bool = False,
):
for field in fields(GroupedGEMMResult):
ref_value = getattr(grouped_result, field.name)
test_value = getattr(fused_result, field.name)
diff = (ref_value - test_value).abs().max()
# second_gemm in torch grouped gemm is not yet unpermuted so comparing the fused unpermuted second_gemm will result in error
# instead the hidden_states_unpermute should match since hidden_states_unpermute for the fused result is the same as second_gemm
if field.name == "second_gemm" and permute_y:
continue
if verbose:
print(f"{field.name} diff: {diff.detach().cpu().item():.6f}")
assert torch.allclose(ref_value, test_value, atol=atol, rtol=rtol), (
f"{field.name} diff: {diff.detach().cpu().item():.6f}"
)
def run_forward(model: nn.Module, X: torch.Tensor, is_grouped_gemm: bool = False):
X = X.detach().clone().requires_grad_(True)
output, router_logits = model(X)
if is_grouped_gemm:
result = ForwardResult(
output=output.hidden_states,
router_logits=router_logits,
X=X,
grouped_gemm_result=output,
)
else:
result = ForwardResult(output=output, router_logits=router_logits, X=X)
return result
def run_backward(
model: nn.Module, grad_output: torch.Tensor, output: torch.Tensor, X: torch.Tensor
):
output.backward(grad_output)
assert X.grad is not None
for name, param in model.named_parameters():
assert param.grad is not None, f"{name} grad is None"
if isinstance(model, Qwen3MoeSparseMoeBlock):
gate_grad = model.gate.weight.grad
gate_proj_grad = torch.stack(
[expert.gate_proj.weight.grad for expert in model.experts]
)
up_proj_grad = torch.stack(
[expert.up_proj.weight.grad for expert in model.experts]
)
down_proj_grad = torch.stack(
[expert.down_proj.weight.grad for expert in model.experts]
)
elif isinstance(model, Qwen3MoeGroupedGEMMBlock):
gate_grad = model.gate.grad
gate_proj_grad, up_proj_grad = model.gate_up_proj.grad.chunk(2, dim=1)
down_proj_grad = model.down_proj.grad
else:
raise ValueError(f"Unsupported model type: {type(model)}")
return BackwardResult(
X_grad=X.grad,
gate_grad=gate_grad,
gate_proj_grad=gate_proj_grad,
up_proj_grad=up_proj_grad,
down_proj_grad=down_proj_grad,
)
class Qwen3MoeFusedGroupedGEMMBlock(Qwen3MoeGroupedGEMMBlock):
"""
Reference implementation of MoE block using grouped gemm.
This is the same as the Qwen3MoeGroupedGEMMBlock but with triton grouped gemm in place of torch-native grouped gemm implementation.
NOTE: This is NOT to be used for production as it contains many extra checks and saves all intermediate results for debugging.
See grouped_gemm/reference/moe_block.py for a cleaner implementation.
"""
def __init__(
self,
config: Qwen3MoeConfig,
gate: torch.Tensor,
gate_up_proj: torch.Tensor,
down_proj: torch.Tensor,
permute_x: bool = False,
permute_y: bool = False,
autotune: bool = True,
kernel_config_fwd: KernelConfigForward = None,
kernel_config_bwd_dW: KernelConfigBackward_dW = None,
kernel_config_bwd_dX: KernelConfigBackward_dX = None,
):
super().__init__(config, gate, gate_up_proj, down_proj)
self.permute_x = permute_x
self.permute_y = permute_y
self.autotune = autotune
if not autotune:
assert (
kernel_config_fwd is not None
and kernel_config_bwd_dW is not None
and kernel_config_bwd_dX is not None
), "Kernel configs must be provided if autotune is False"
self.kernel_config_fwd = kernel_config_fwd
self.kernel_config_bwd_dW = kernel_config_bwd_dW
self.kernel_config_bwd_dX = kernel_config_bwd_dX
@classmethod
def from_hf(
cls,
moe_block: Qwen3MoeSparseMoeBlock,
permute_x: bool = False,
permute_y: bool = False,
autotune: bool = True,
kernel_config_fwd: KernelConfigForward = None,
kernel_config_bwd_dW: KernelConfigBackward_dW = None,
kernel_config_bwd_dX: KernelConfigBackward_dX = None,
):
config: Qwen3MoeConfig = moe_block.experts[0].config
gate, gate_up_proj, down_proj = Qwen3MoeGroupedGEMMBlock.extract_hf_weights(
moe_block
)
return cls(
config,
gate,
gate_up_proj,
down_proj,
permute_x=permute_x,
permute_y=permute_y,
autotune=autotune,
kernel_config_fwd=kernel_config_fwd,
kernel_config_bwd_dW=kernel_config_bwd_dW,
kernel_config_bwd_dX=kernel_config_bwd_dX,
)
def forward(self, hidden_states: torch.Tensor, debug: bool = False) -> torch.Tensor:
batch_size, sequence_length, hidden_dim = hidden_states.shape
num_tokens = batch_size * sequence_length
total_tokens = num_tokens * self.top_k
hidden_states = hidden_states.view(-1, hidden_dim)
router_logits, routing_weights, selected_experts = self.run_router(
hidden_states
)
# Pre-processing
# 1. Compute tokens per expert and indices for gathering tokes from token order to expert order
# NOTE: these are auxiliary data structs which don't need to be recorded in autograd graph
token_counts_by_expert, gather_indices = (
self.get_token_counts_and_gather_indices(selected_experts)
)
# 2. permute_x -> permutation will be fused in prologue of first grouped gemm
if not self.permute_x:
hidden_states = permute(hidden_states, gather_indices, self.top_k)
assert hidden_states.shape == (total_tokens, hidden_dim)
# Start expert computation
first_gemm = grouped_gemm(
X=hidden_states,
W=self.gate_up_proj,
m_sizes=token_counts_by_expert,
gather_indices=gather_indices,
topk=self.top_k,
permute_x=self.permute_x,
permute_y=False, # output of first grouped gemm should never be permuted
autotune=self.autotune,
kernel_config_fwd=self.kernel_config_fwd,
kernel_config_bwd_dW=self.kernel_config_bwd_dW,
kernel_config_bwd_dX=self.kernel_config_bwd_dX,
is_first_gemm=True,
)
assert first_gemm.shape == (total_tokens, 2 * self.moe_intermediate_size)
intermediate = self.act_and_mul(first_gemm)
assert intermediate.shape == (total_tokens, self.moe_intermediate_size)
second_gemm = grouped_gemm(
X=intermediate,
W=self.down_proj,
m_sizes=token_counts_by_expert,
gather_indices=gather_indices,
topk=self.top_k,
permute_x=False,
permute_y=self.permute_y,
autotune=self.autotune,
kernel_config_fwd=self.kernel_config_fwd,
kernel_config_bwd_dW=self.kernel_config_bwd_dW,
kernel_config_bwd_dX=self.kernel_config_bwd_dX,
is_first_gemm=False,
)
assert second_gemm.shape == (total_tokens, hidden_dim)
# Post-processing
# 1. Unpermute from expert order to token order
if not self.permute_y:
hidden_states_unpermute = unpermute(second_gemm, gather_indices)
assert hidden_states_unpermute.shape == (total_tokens, hidden_dim)
else:
hidden_states_unpermute = second_gemm
# 2. Merge topk weights
hidden_states = (
hidden_states_unpermute.view(num_tokens, self.top_k, hidden_dim)
* routing_weights[..., None]
)
hidden_states = hidden_states.sum(dim=1)
assert hidden_states.shape == (num_tokens, hidden_dim)
hidden_states = hidden_states.view(batch_size, sequence_length, hidden_dim)
return GroupedGEMMResult(
token_counts_by_expert=token_counts_by_expert,
gather_indices=gather_indices,
topk_weights=routing_weights,
first_gemm=first_gemm,
intermediate=intermediate,
second_gemm=second_gemm,
hidden_states_unpermute=hidden_states_unpermute,
hidden_states=hidden_states,
), router_logits

View file

@ -0,0 +1,35 @@
#!/bin/bash
set -euo pipefail
SEQLENS=(1024)
DTYPES=(bfloat16)
PERMUTE_X=(false true)
PERMUTE_Y=(false true)
AUTOTUNE=(false true)
for SEQLEN in "${SEQLENS[@]}"; do
for DTYPE in "${DTYPES[@]}"; do
for PX in "${PERMUTE_X[@]}"; do
for PY in "${PERMUTE_Y[@]}"; do
for AT in "${AUTOTUNE[@]}"; do
ARGS=()
[[ "$PX" == "true" ]] && ARGS+=("--permute_x")
[[ "$PY" == "true" ]] && ARGS+=("--permute_y")
[[ "$AT" == "true" ]] && ARGS+=("--autotune")
ARGS+=(--seqlen "$SEQLEN" --dtype "$DTYPE")
echo "Running with args: ${ARGS[*]}"
if ! python -m tests.test_qwen3_moe "${ARGS[@]}"; then
echo "❌ Test failed with args: --permute_x=$PX --permute_y=$PY --autotune=$AT --seqlen=$SEQLEN --dtype=$DTYPE" >&2
else
echo "✅ Test passed with args: --permute_x=$PX --permute_y=$PY --autotune=$AT --seqlen=$SEQLEN --dtype=$DTYPE"
fi
done
done
done
done
done

File diff suppressed because it is too large Load diff

View file

@ -0,0 +1,267 @@
import argparse
from contextlib import contextmanager
import pytest
import torch
from transformers import AutoConfig
from transformers.models.qwen3_moe import Qwen3MoeConfig
from transformers.models.qwen3_moe.modeling_qwen3_moe import Qwen3MoeSparseMoeBlock
from grouped_gemm.kernels.tuning import (
KernelConfigBackward_dW,
KernelConfigBackward_dX,
KernelConfigForward,
)
from grouped_gemm.reference.moe_ops import Qwen3MoeGroupedGEMMBlock
from .moe_utils import (
Qwen3MoeFusedGroupedGEMMBlock,
check_fwd,
check_grads,
check_grouped_gemm_results,
run_backward,
run_forward,
)
"""
Qwen3 MoE tests
NOTE: Test this as a module and NOT with pytest as running with pytest results in random numerical errors: python -m tests.test_qwen3_moe --permute_x --permute_y --autotune NOT pytest -sv tests/test_qwen3_moe.py
More specifically, all tests pass when run individually, but some will fail randomly (even with the same seed) when the entire test is run as a parametrized test suite using pytest, likely due to how pytest interacts with triton / autotuning.
See tests/run_qwen3_moe_tests.sh for a script that runs all the tests
The tests run the following:
Huggingface's Qwen3 MoE block (Qwen3MoeSparseMoeBlock)
Torch-native grouped gemm version of MoE block (Qwen3MoeGroupedGEMMBlock), which is the HF block with the expert computation replaced with a torch-native grouped gemm
Triton kernel grouped gemm version of MoE block (Qwen3MoeFusedGroupedGEMMBlock), which is the HF block with the expert computation replaced with the fused triton grouped gemm kernel
The tests check the following:
- HF MoE block vs torch grouped gemm MoE block (sanity check)
- torch grouped gemm MoE block vs fused grouped gemm MoE block -- this allows us to test each of the intermediate results for easier debugging
- HF MoE block vs fused grouped gemm MoE block -- this is the actual test
Both forward and backward passes are tests:
- forward: output of the moe block
- backwards:
- X: gradient of the input to the moe block
- gate.weight: gradient of the gate weights (router weights)
- gate_proj: gradient of concatenated gate projections
- up_proj: gradient of the concatenated up projections
- down_proj: gradient of the concatenated down projections
Additionally, for the torch grouped gemm and triton grouped gemm versions, the intermediate outputs of the forward pass are checked:
- first_gemm: output of the first grouped gemm (X @ fused_gate_proj)
- intermediate: output of silu_mul(first_gemm)
- second_gemm: output of the second grouped gemm (intermediate @ down_proj)
- hidden_states_unpermute: output of the second_gemm after unpermuting back to token order (from expert grouped order); in the case where the permutation is fused in the triton kernel, this is the same as second_gemm
- hidden_states: output with the topk_weights applied
"""
TOLERANCES = {
torch.bfloat16: (1e-2, 1e-2),
torch.float16: (1e-3, 1e-3),
torch.float: (1e-5, 1e-5),
}
@pytest.fixture(scope="module")
def model_id():
return "Qwen/Qwen3-30B-A3B"
@pytest.fixture(scope="module")
def config(model_id: str):
return AutoConfig.from_pretrained(model_id)
@contextmanager
def test_context(prelude, epilogue="Passed!", char="-", num_chars=80):
print(char * num_chars)
print(prelude)
yield
print(epilogue)
print(char * num_chars)
SEED = 42
SEQ_LENS = [1024]
DTYPES = [torch.bfloat16]
# Reduce the number of autotuning configs to prevent excessive runtime
NUM_AUTOTUNE_CONFIGS = 50
@pytest.mark.parametrize(
"permute_y", [True], ids=lambda x: "permute_y" if x else "no_permute_y"
)
@pytest.mark.parametrize(
"permute_x", [True], ids=lambda x: "permute_x" if x else "no_permute_x"
)
@pytest.mark.parametrize(
"autotune", [True], ids=lambda x: "autotune" if x else "manual"
)
@pytest.mark.parametrize("seqlen", SEQ_LENS, ids=lambda x: f"seqlen={x}")
@pytest.mark.parametrize("dtype", DTYPES, ids=str)
def test_qwen3_moe(
config: Qwen3MoeConfig,
seqlen: int,
dtype: torch.dtype,
permute_x: bool,
permute_y: bool,
autotune: bool,
atol: float,
rtol: float,
):
torch.manual_seed(
SEED
) # Should not be needed when running using pytest -- autouse fixture in conftest.py
device = "cuda"
hidden_size = config.hidden_size
bs = 1
# Reference op -- HF
moe_block = Qwen3MoeSparseMoeBlock(config).to(device, dtype)
# Torch-native grouped gemm version of MoE Block -- for sanity checking
grouped_gemm_block = Qwen3MoeGroupedGEMMBlock.from_hf(moe_block).to(device, dtype)
grouped_gemm_block.check_weights(moe_block)
if not autotune:
kernel_config_fwd = KernelConfigForward()
kernel_config_bwd_dW = KernelConfigBackward_dW()
kernel_config_bwd_dX = KernelConfigBackward_dX()
else:
from grouped_gemm.kernels.backward import (
_autotuned_grouped_gemm_dW_kernel,
_autotuned_grouped_gemm_dX_kernel,
)
from grouped_gemm.kernels.forward import _autotuned_grouped_gemm_forward_kernel
# Hack to reduce number of autotuning configs
_autotuned_grouped_gemm_forward_kernel.configs = (
_autotuned_grouped_gemm_forward_kernel.configs[:NUM_AUTOTUNE_CONFIGS]
)
_autotuned_grouped_gemm_dW_kernel.configs = (
_autotuned_grouped_gemm_dW_kernel.configs[:NUM_AUTOTUNE_CONFIGS]
)
_autotuned_grouped_gemm_dX_kernel.configs = (
_autotuned_grouped_gemm_dX_kernel.configs[:NUM_AUTOTUNE_CONFIGS]
)
kernel_config_fwd = None
kernel_config_bwd_dW = None
kernel_config_bwd_dX = None
# Triton kernel grouped gemm version of MoE Block -- this is what we're testing
fused_gemm_block = Qwen3MoeFusedGroupedGEMMBlock.from_hf(
moe_block,
permute_x=permute_x,
permute_y=permute_y,
autotune=autotune,
kernel_config_fwd=kernel_config_fwd,
kernel_config_bwd_dW=kernel_config_bwd_dW,
kernel_config_bwd_dX=kernel_config_bwd_dX,
).to(device, dtype)
fused_gemm_block.check_weights(moe_block)
X = torch.randn(
bs, seqlen, hidden_size, dtype=dtype, device=device, requires_grad=True
)
# Forward
ref_result = run_forward(moe_block, X, is_grouped_gemm=False)
grouped_result = run_forward(grouped_gemm_block, X, is_grouped_gemm=True)
fused_result = run_forward(fused_gemm_block, X, is_grouped_gemm=True)
with test_context(
"Testing forward pass",
epilogue="Passed forward tests!",
char="=",
num_chars=100,
):
# Sanity checks
with test_context("Checking HF vs torch grouped gemm MoE forward outputs..."):
check_fwd(ref_result, grouped_result, atol, rtol, verbose=False)
with test_context(
"Checking torch grouped gemm MoE vs fused grouped gemm MoE forward outputs..."
):
# We implement a custom check for grouped gemm results to test each of the intermediate results for easier debugging
check_grouped_gemm_results(
grouped_result.grouped_gemm_result,
fused_result.grouped_gemm_result,
permute_y=permute_y,
atol=atol,
rtol=rtol,
verbose=False,
)
# Actual test
with test_context("Checking HF vs fused grouped gemm MoE forward outputs..."):
check_fwd(ref_result, fused_result, atol, rtol, verbose=True)
# Backward
grad_output = torch.randn_like(ref_result.output)
ref_backward_result = run_backward(
moe_block, grad_output, output=ref_result.output, X=ref_result.X
)
grouped_backward_result = run_backward(
grouped_gemm_block,
grad_output,
output=grouped_result.output,
X=grouped_result.X,
)
fused_backward_result = run_backward(
fused_gemm_block, grad_output, output=fused_result.output, X=fused_result.X
)
with test_context(
"Testing backward pass",
epilogue="Passed backward tests!",
char="=",
num_chars=100,
):
# Sanity checks
with test_context("Checking HF vs torch grouped gemm MoE grads..."):
check_grads(
ref_backward_result, grouped_backward_result, atol, rtol, verbose=False
)
with test_context(
"Checking torch grouped gemm MoE vs fused grouped gemm MoE grads..."
):
check_grads(
grouped_backward_result,
fused_backward_result,
atol,
rtol,
verbose=False,
)
# Actual test
with test_context("Checking HF vs fused grouped gemm MoE grads..."):
check_grads(
ref_backward_result, fused_backward_result, atol, rtol, verbose=True
)
if __name__ == "__main__":
parser = argparse.ArgumentParser()
parser.add_argument("--seqlen", type=int, default=1024)
parser.add_argument(
"--dtype", type=str, choices=["bfloat16", "float16"], default="bfloat16"
)
parser.add_argument("--permute_x", action="store_true")
parser.add_argument("--permute_y", action="store_true")
parser.add_argument("--autotune", action="store_true")
args = parser.parse_args()
args.dtype = getattr(torch, args.dtype)
args_dict = vars(args)
model_id = "Qwen/Qwen3-30B-A3B"
config = AutoConfig.from_pretrained(model_id)
atol, rtol = TOLERANCES[args.dtype]
print(
f"Testing {model_id} with seqlen={args.seqlen}, dtype={args.dtype}, permute_x={args.permute_x}, permute_y={args.permute_y}, autotune={args.autotune}, atol={atol}, rtol={rtol}"
)
test_qwen3_moe(config, atol=atol, rtol=rtol, **args_dict)